@@ -2,9 +2,9 @@
  * dpkg-deb - construction and deconstruction of *.deb archives
  * build.c - building archives
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
  * Copyright © 2000,2001 Wichert Akkerman <wakkerma@debian.org>
- * Copyright © 2007-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2007-2015 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,7 +29,6 @@
 
 #include <errno.h>
 #include <limits.h>
-#include <ctype.h>
 #include <string.h>
 #include <dirent.h>
 #include <fcntl.h>
@@ -40,6 +39,7 @@
 #include <stdio.h>
 
 #include <dpkg/i18n.h>
+#include <dpkg/c-ctype.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
 #include <dpkg/path.h>
@@ -121,8 +121,6 @@ file_info_get(const char *root, int fd)
       break;
 
     varbuf_trunc(&fn, fn.used + 1);
-    if (fn.used >= MAXFILENAME)
-      ohshit(_("file name '%.50s...' is too long"), fn.buf + root_len);
   }
 
   fi = file_info_new(fn.buf + root_len);
@@ -185,7 +183,7 @@ file_treewalk_feed(const char *dir, int fd_out)
     close(pipefd[1]);
 
     if (chdir(dir))
-      ohshite(_("failed to chdir to `%.255s'"), dir);
+      ohshite(_("failed to chdir to '%.255s'"), dir);
 
     execlp(FIND, "find", ".", "-path", "./" BUILDCONTROLDIR, "-prune", "-o",
            "-print0", NULL);
@@ -196,6 +194,8 @@ file_treewalk_feed(const char *dir, int fd_out)
   /* We need to reorder the files so we can make sure that symlinks
    * will not appear before their target. */
   while ((fi = file_info_get(dir, pipefd[0])) != NULL) {
+    if (strchr(fi->fn, '\n'))
+      ohshit(_("newline not allowed in pathname '%s'"), fi->fn);
     if (S_ISLNK(fi->st.st_mode)) {
       file_info_list_append(&symlist, &symlist_end, fi);
     } else {
@@ -207,7 +207,7 @@ file_treewalk_feed(const char *dir, int fd_out)
   }
 
   close(pipefd[0]);
-  subproc_wait_check(pid, "find", 0);
+  subproc_reap(pid, "find", 0);
 
   for (fi = symlist; fi; fi = fi->next)
     if (fd_write(fd_out, fi->fn, strlen(fi->fn) + 1) < 0)
@@ -228,13 +228,13 @@ static const char *const maintainerscripts[] = {
  * Check control directory and file permissions.
  */
 static void
-check_file_perms(const char *dir)
+check_file_perms(const char *ctrldir)
 {
   struct varbuf path = VARBUF_INIT;
   const char *const *mscriptp;
   struct stat mscriptstab;
 
-  varbuf_printf(&path, "%s/%s/", dir, BUILDCONTROLDIR);
+  varbuf_printf(&path, "%s/", ctrldir);
   if (lstat(path.buf, &mscriptstab))
     ohshite(_("unable to stat control directory"));
   if (!S_ISDIR(mscriptstab.st_mode))
@@ -246,19 +246,19 @@ check_file_perms(const char *dir)
 
   for (mscriptp = maintainerscripts; *mscriptp; mscriptp++) {
     varbuf_reset(&path);
-    varbuf_printf(&path, "%s/%s/%s", dir, BUILDCONTROLDIR, *mscriptp);
+    varbuf_printf(&path, "%s/%s", ctrldir, *mscriptp);
     if (!lstat(path.buf, &mscriptstab)) {
       if (S_ISLNK(mscriptstab.st_mode))
         continue;
       if (!S_ISREG(mscriptstab.st_mode))
-        ohshit(_("maintainer script `%.50s' is not a plain file or symlink"),
+        ohshit(_("maintainer script '%.50s' is not a plain file or symlink"),
                *mscriptp);
       if ((mscriptstab.st_mode & 07557) != 0555)
-        ohshit(_("maintainer script `%.50s' has bad permissions %03lo "
+        ohshit(_("maintainer script '%.50s' has bad permissions %03lo "
                  "(must be >=0555 and <=0775)"),
                *mscriptp, (unsigned long)(mscriptstab.st_mode & 07777));
     } else if (errno != ENOENT) {
-      ohshite(_("maintainer script `%.50s' is not stattable"), *mscriptp);
+      ohshite(_("maintainer script '%.50s' is not stattable"), *mscriptp);
     }
   }
 
@@ -269,7 +269,7 @@ check_file_perms(const char *dir)
  * Check if conffiles contains sane information.
  */
 static void
-check_conffiles(const char *dir)
+check_conffiles(const char *ctrldir, const char *rootdir)
 {
   FILE *cf;
   struct varbuf controlfile = VARBUF_INIT;
@@ -277,7 +277,7 @@ check_conffiles(const char *dir)
   struct file_info *conffiles_head = NULL;
   struct file_info *conffiles_tail = NULL;
 
-  varbuf_printf(&controlfile, "%s/%s/%s", dir, BUILDCONTROLDIR, CONFFILESFILE);
+  varbuf_printf(&controlfile, "%s/%s", ctrldir, CONFFILESFILE);
 
   cf = fopen(controlfile.buf, "r");
   if (cf == NULL) {
@@ -295,27 +295,21 @@ check_conffiles(const char *dir)
     if (!n)
       ohshite(_("empty string from fgets reading conffiles"));
 
-    if (conffilename[n - 1] != '\n') {
-      int c;
-
-      warning(_("conffile name '%.50s...' is too long, or missing final newline"),
-              conffilename);
-      while ((c = getc(cf)) != EOF && c != '\n');
-
-      continue;
-    }
+    if (conffilename[n - 1] != '\n')
+      ohshit(_("conffile name '%s' is too long, or missing final newline"),
+             conffilename);
 
     conffilename[n - 1] = '\0';
     varbuf_reset(&controlfile);
-    varbuf_printf(&controlfile, "%s/%s", dir, conffilename);
+    varbuf_printf(&controlfile, "%s/%s", rootdir, conffilename);
     if (lstat(controlfile.buf, &controlstab)) {
       if (errno == ENOENT) {
-        if ((n > 1) && isspace(conffilename[n - 2]))
+        if ((n > 1) && c_isspace(conffilename[n - 2]))
           warning(_("conffile filename '%s' contains trailing white spaces"),
                   conffilename);
-        ohshit(_("conffile `%.250s' does not appear in package"), conffilename);
+        ohshit(_("conffile '%.250s' does not appear in package"), conffilename);
       } else
-        ohshite(_("conffile `%.250s' is not stattable"), conffilename);
+        ohshite(_("conffile '%.250s' is not stattable"), conffilename);
     } else if (!S_ISREG(controlstab.st_mode)) {
       warning(_("conffile '%s' is not a plain file"), conffilename);
     }
@@ -338,98 +332,168 @@ check_conffiles(const char *dir)
   fclose(cf);
 }
 
-static const char *arbitrary_fields[] = {
-  "Built-For-Profiles",
-  "Built-Using",
-  "Package-Type",
-  "Subarchitecture",
-  "Kernel-Version",
-  "Installer-Menu-Item",
-  "Homepage",
-  "Tag",
-  NULL
-};
-
-static const char private_prefix[] = "Private-";
-
-static bool
-known_arbitrary_field(const struct arbitraryfield *field)
-{
-  const char **known;
-
-  /* Always accept fields starting with a private field prefix. */
-  if (strncasecmp(field->name, private_prefix, strlen(private_prefix)) == 0)
-    return true;
-
-  for (known = arbitrary_fields; *known; known++)
-    if (strcasecmp(field->name, *known) == 0)
-      return true;
-
-  return false;
-}
-
 /**
- * Perform some sanity checks on the to-be-built package.
+ * Check the control file.
  *
- * @return The pkginfo struct from the parsed control file.
+ * @param dir	The directory from where to build the binary package.
+ * @return	The pkginfo struct from the parsed control file.
  */
 static struct pkginfo *
-check_new_pkg(const char *dir)
+check_control_file(const char *ctrldir)
 {
   struct pkginfo *pkg;
-  struct arbitraryfield *field;
   char *controlfile;
-  int warns;
 
-  /* Start by reading in the control file so we can check its contents. */
-  m_asprintf(&controlfile, "%s/%s/%s", dir, BUILDCONTROLDIR, CONTROLFILE);
+  controlfile = str_fmt("%s/%s", ctrldir, CONTROLFILE);
   parsedb(controlfile, pdb_parse_binary, &pkg);
 
   if (strspn(pkg->set->name, "abcdefghijklmnopqrstuvwxyz0123456789+-.") !=
       strlen(pkg->set->name))
-    ohshit(_("package name has characters that aren't lowercase alphanums or `-+.'"));
-  if (pkg->priority == pri_other)
+    ohshit(_("package name has characters that aren't lowercase alphanums or '-+.'"));
+  if (pkg->priority == PKG_PRIO_OTHER)
     warning(_("'%s' contains user-defined Priority value '%s'"),
             controlfile, pkg->otherpriority);
-  for (field = pkg->available.arbs; field; field = field->next) {
-    if (known_arbitrary_field(field))
-      continue;
-
-    warning(_("'%s' contains user-defined field '%s'"), controlfile,
-            field->name);
-  }
 
   free(controlfile);
 
-  check_file_perms(dir);
-  check_conffiles(dir);
+  return pkg;
+}
+
+/**
+ * Perform some sanity checks on the to-be-built package control area.
+ *
+ * @param dir	The directory from where to build the binary package.
+ * @return	The pkginfo struct from the parsed control file.
+ */
+static struct pkginfo *
+check_control_area(const char *ctrldir, const char *rootdir)
+{
+  struct pkginfo *pkg;
+  int warns;
+
+  /* Start by reading in the control file so we can check its contents. */
+  pkg = check_control_file(ctrldir);
+  check_file_perms(ctrldir);
+  check_conffiles(ctrldir, rootdir);
 
   warns = warning_get_count();
   if (warns)
-    warning(P_("ignoring %d warning about the control file(s)\n",
-               "ignoring %d warnings about the control file(s)\n", warns),
+    warning(P_("ignoring %d warning about the control file(s)",
+               "ignoring %d warnings about the control file(s)", warns),
             warns);
 
   return pkg;
 }
 
 /**
- * Generate the pathname for the to-be-built package.
+ * Generate the pathname for the destination binary package.
+ *
+ * If the pathname cannot be computed, because the destination is a directory,
+ * then NULL will be returned.
+ *
+ * @param dir	The directory from where to build the binary package.
+ * @param dest	The destination name, either a file or directory name.
+ * @return	The pathname for the package being built.
+ */
+static char *
+gen_dest_pathname(const char *dir, const char *dest)
+{
+  if (dest) {
+    struct stat dest_stab;
+
+    if (stat(dest, &dest_stab)) {
+      if (errno != ENOENT)
+        ohshite(_("unable to check for existence of archive '%.250s'"), dest);
+    } else if (S_ISDIR(dest_stab.st_mode)) {
+      /* Need to compute the destination name from the package control file. */
+      return NULL;
+    }
+
+    return m_strdup(dest);
+  } else {
+    char *pathname;
+
+    pathname = m_malloc(strlen(dir) + sizeof(DEBEXT));
+    strcpy(pathname, dir);
+    path_trim_slash_slashdot(pathname);
+    strcat(pathname, DEBEXT);
+
+    return pathname;
+  }
+}
+
+/**
+ * Generate the pathname for the destination binary package from control file.
  *
  * @return The pathname for the package being built.
  */
 static char *
-pkg_get_pathname(const char *dir, struct pkginfo *pkg)
+gen_dest_pathname_from_pkg(const char *dir, struct pkginfo *pkg)
 {
-  char *path;
-  const char *versionstring, *arch_sep;
+  const char *arch_sep;
 
-  versionstring = versiondescribe(&pkg->available.version, vdew_never);
-  arch_sep = pkg->available.arch->type == arch_none ? "" : "_";
-  m_asprintf(&path, "%s/%s_%s%s%s%s", dir, pkg->set->name, versionstring,
-             arch_sep, pkg->available.arch->name, DEBEXT);
+  if (pkg->available.arch->type == DPKG_ARCH_NONE ||
+      pkg->available.arch->type == DPKG_ARCH_EMPTY)
+    arch_sep = "";
+  else
+    arch_sep = "_";
 
-  return path;
+  return str_fmt("%s/%s_%s%s%s%s", dir, pkg->set->name,
+                 versiondescribe(&pkg->available.version, vdew_never),
+                 arch_sep, pkg->available.arch->name, DEBEXT);
+}
+
+typedef void filenames_feed_func(const char *dir, int fd_out);
+
+/**
+ * Pack the contents of a directory into a tarball.
+ */
+static void
+tarball_pack(const char *dir, filenames_feed_func *tar_filenames_feeder,
+             struct compress_params *tar_compress_params, int fd_out)
+{
+  int pipe_filenames[2], pipe_tarball[2];
+  pid_t pid_tar, pid_comp;
+
+  /* Fork off a tar. We will feed it a list of filenames on stdin later. */
+  m_pipe(pipe_filenames);
+  m_pipe(pipe_tarball);
+  pid_tar = subproc_fork();
+  if (pid_tar == 0) {
+    m_dup2(pipe_filenames[0], 0);
+    close(pipe_filenames[0]);
+    close(pipe_filenames[1]);
+    m_dup2(pipe_tarball[1], 1);
+    close(pipe_tarball[0]);
+    close(pipe_tarball[1]);
+
+    if (chdir(dir))
+      ohshite(_("failed to chdir to '%.255s'"), dir);
+
+    execlp(TAR, "tar", "-cf", "-", "--format=gnu", "--null", "--no-unquote",
+                       "--no-recursion", "-T", "-", NULL);
+    ohshite(_("unable to execute %s (%s)"), "tar -cf", TAR);
+  }
+  close(pipe_filenames[0]);
+  close(pipe_tarball[1]);
+
+  /* Of course we should not forget to compress the archive as well. */
+  pid_comp = subproc_fork();
+  if (pid_comp == 0) {
+    close(pipe_filenames[1]);
+    compress_filter(tar_compress_params, pipe_tarball[0], fd_out,
+                    _("compressing tar member"));
+    exit(0);
+  }
+  close(pipe_tarball[0]);
+
+  /* All the pipes are set, now lets start feeding filenames to tar. */
+  tar_filenames_feeder(dir, pipe_filenames[1]);
+
+  /* All done, clean up wait for tar and <compress> to finish their job. */
+  close(pipe_filenames[1]);
+  subproc_reap(pid_comp, _("<compress> from tar -cf"), 0);
+  subproc_reap(pid_tar, "tar -cf", 0);
 }
 
 /**
@@ -438,55 +502,41 @@ pkg_get_pathname(const char *dir, struct pkginfo *pkg)
 int
 do_build(const char *const *argv)
 {
+  struct compress_params control_compress_params;
   struct dpkg_error err;
-  const char *debar, *dir;
-  bool subdir;
+  const char *dir, *dest;
+  char *ctrldir;
+  char *debar;
   char *tfbuf;
   int arfd;
-  int p1[2], p2[2], gzfd;
+  int p1[2], gzfd;
   pid_t c1, c2;
 
   /* Decode our arguments. */
   dir = *argv++;
   if (!dir)
     badusage(_("--%s needs a <directory> argument"), cipaction->olong);
-  subdir = false;
-  debar = *argv++;
-  if (debar != NULL) {
-    struct stat debarstab;
 
-    if (*argv)
-      badusage(_("--%s takes at most two arguments"), cipaction->olong);
+  dest = *argv++;
+  if (dest && *argv)
+    badusage(_("--%s takes at most two arguments"), cipaction->olong);
 
-    if (stat(debar, &debarstab)) {
-      if (errno != ENOENT)
-        ohshite(_("unable to check for existence of archive `%.250s'"), debar);
-    } else if (S_ISDIR(debarstab.st_mode)) {
-      subdir = true;
-    }
-  } else {
-    char *m;
-
-    m= m_malloc(strlen(dir) + sizeof(DEBEXT));
-    strcpy(m, dir);
-    path_trim_slash_slashdot(m);
-    strcat(m, DEBEXT);
-    debar= m;
-  }
+  debar = gen_dest_pathname(dir, dest);
+  ctrldir = str_fmt("%s/%s", dir, BUILDCONTROLDIR);
 
   /* Perform some sanity checks on the to-be-build package. */
   if (nocheckflag) {
-    if (subdir)
+    if (debar == NULL)
       ohshit(_("target is directory - cannot skip control file check"));
     warning(_("not checking contents of control area"));
     printf(_("dpkg-deb: building an unknown package in '%s'.\n"), debar);
   } else {
     struct pkginfo *pkg;
 
-    pkg = check_new_pkg(dir);
-    if (subdir)
-      debar = pkg_get_pathname(debar, pkg);
-    printf(_("dpkg-deb: building package `%s' in `%s'.\n"),
+    pkg = check_control_area(ctrldir, dir);
+    if (debar == NULL)
+      debar = gen_dest_pathname_from_pkg(dest, pkg);
+    printf(_("dpkg-deb: building package '%s' in '%s'.\n"),
            pkg->set->name, debar);
   }
   m_output(stdout, _("<standard output>"));
@@ -495,21 +545,21 @@ do_build(const char *const *argv)
    * build something. Let's start by making the ar-wrapper. */
   arfd = creat(debar, 0644);
   if (arfd < 0)
-    ohshite(_("unable to create `%.255s'"), debar);
+    ohshite(_("unable to create '%.255s'"), debar);
   /* Fork a tar to package the control-section of the package. */
   unsetenv("TAR_OPTIONS");
   m_pipe(p1);
   c1 = subproc_fork();
   if (!c1) {
     m_dup2(p1[1],1); close(p1[0]); close(p1[1]);
-    if (chdir(dir))
-      ohshite(_("failed to chdir to `%.255s'"), dir);
-    if (chdir(BUILDCONTROLDIR))
-      ohshite(_("failed to chdir to `%.255s'"), ".../DEBIAN");
+    if (chdir(ctrldir))
+      ohshite(_("failed to chdir to '%.255s'"), ctrldir);
     execlp(TAR, "tar", "-cf", "-", "--format=gnu", ".", NULL);
     ohshite(_("unable to execute %s (%s)"), "tar -cf", TAR);
   }
   close(p1[1]);
+  free(ctrldir);
+
   /* Create a temporary file to store the control data in. Immediately
    * unlink our temporary file so others can't mess with it. */
   tfbuf = path_make_temp_template("dpkg-deb");
@@ -522,21 +572,25 @@ do_build(const char *const *argv)
            tfbuf);
   free(tfbuf);
 
-  /* And run gzip to compress our control archive. */
+  /* And run the compressor on our control archive. */
+  if (opt_uniform_compression) {
+    control_compress_params = compress_params;
+  } else {
+    control_compress_params.type = COMPRESSOR_TYPE_GZIP;
+    control_compress_params.strategy = COMPRESSOR_STRATEGY_NONE;
+    control_compress_params.level = -1;
+    if (!compressor_check_params(&control_compress_params, &err))
+      internerr("invalid control member compressor params: %s", err.str);
+  }
+
   c2 = subproc_fork();
   if (!c2) {
-    struct compress_params params;
-
-    params.type = compressor_type_gzip;
-    params.strategy = compressor_strategy_none;
-    params.level = 9;
-
-    compress_filter(&params, p1[0], gzfd, _("compressing control member"));
+    compress_filter(&control_compress_params, p1[0], gzfd, _("compressing control member"));
     exit(0);
   }
   close(p1[0]);
-  subproc_wait_check(c2, "gzip -9c", 0);
-  subproc_wait_check(c1, "tar -cf", 0);
+  subproc_reap(c2, _("<compress> from tar -cf"), 0);
+  subproc_reap(c1, "tar -cf", 0);
 
   if (lseek(gzfd, 0, SEEK_SET))
     ohshite(_("failed to rewind temporary file (%s)"), _("control member"));
@@ -552,7 +606,7 @@ do_build(const char *const *argv)
     sprintf(versionbuf, "%-8s\n%jd\n", OLDARCHIVEVERSION,
             (intmax_t)controlstab.st_size);
     if (fd_write(arfd, versionbuf, strlen(versionbuf)) < 0)
-      ohshite(_("error writing `%s'"), debar);
+      ohshite(_("error writing '%s'"), debar);
     if (fd_fd_copy(gzfd, arfd, -1, &err) < 0)
       ohshit(_("cannot copy '%s' into archive '%s': %s"), _("control member"),
              debar, err.str);
@@ -561,7 +615,7 @@ do_build(const char *const *argv)
     char adminmember[16 + 1];
 
     sprintf(adminmember, "%s%s", ADMINMEMBER,
-            compressor_get_extension(compressor_type_gzip));
+            compressor_get_extension(control_compress_params.type));
 
     dpkg_ar_put_magic(debar, arfd);
     dpkg_ar_member_put_mem(debar, arfd, DEBMAGIC, deb_magic, strlen(deb_magic));
@@ -593,37 +647,10 @@ do_build(const char *const *argv)
   } else {
     internerr("unknown deb format version %d.%d", deb_format.major, deb_format.minor);
   }
-  /* Fork off a tar. We will feed it a list of filenames on stdin later. */
-  m_pipe(p1);
-  m_pipe(p2);
-  c1 = subproc_fork();
-  if (!c1) {
-    m_dup2(p1[0],0); close(p1[0]); close(p1[1]);
-    m_dup2(p2[1],1); close(p2[0]); close(p2[1]);
-    if (chdir(dir))
-      ohshite(_("failed to chdir to `%.255s'"), dir);
-    execlp(TAR, "tar", "-cf", "-", "--format=gnu", "--null", "-T", "-", "--no-recursion", NULL);
-    ohshite(_("unable to execute %s (%s)"), "tar -cf", TAR);
-  }
-  close(p1[0]);
-  close(p2[1]);
-  /* Of course we should not forget to compress the archive as well. */
-  c2 = subproc_fork();
-  if (!c2) {
-    close(p1[1]);
-    compress_filter(&compress_params, p2[0], gzfd, _("compressing data member"));
-    exit(0);
-  }
-  close(p2[0]);
 
-  /* All the pipes are set, now lets walk the tree, and start feeding
-   * filenames to tar. */
-  file_treewalk_feed(dir, p1[1]);
+  /* Pack the directory into a tarball, feeding files from the callback. */
+  tarball_pack(dir, file_treewalk_feed, &compress_params, gzfd);
 
-  /* All done, clean up wait for tar and gzip to finish their job. */
-  close(p1[1]);
-  subproc_wait_check(c2, _("<compress> from tar -cf"), 0);
-  subproc_wait_check(c1, "tar -cf", 0);
   /* Okay, we have data.tar as well now, add it to the ar wrapper. */
   if (deb_format.major == 2) {
     char datamember[16 + 1];
@@ -643,5 +670,7 @@ do_build(const char *const *argv)
   if (close(arfd))
     ohshite(_("unable to close file '%s'"), debar);
 
+  free(debar);
+
   return 0;
 }
