@@ -2,7 +2,7 @@
  * dpkg-deb - construction and deconstruction of *.deb archives
  * extract.c - extracting archives
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
  * Copyright © 2006-2014 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
@@ -187,6 +187,7 @@ extracthalf(const char *debar, const char *dir,
           decompressor = compressor_find_by_extension(extension);
           if (decompressor != COMPRESSOR_TYPE_NONE &&
               decompressor != COMPRESSOR_TYPE_GZIP &&
+              decompressor != COMPRESSOR_TYPE_ZSTD &&
               decompressor != COMPRESSOR_TYPE_XZ)
             ohshit(_("archive '%s' uses unknown compression for member '%.*s', "
                      "giving up"),
@@ -491,7 +492,7 @@ do_raw_extract(const char *const *argv)
     badusage(_("--%s takes at most two arguments (.deb and directory)"),
              cipaction->olong);
 
-  m_asprintf(&control_dir, "%s/%s", dir, EXTRACTCONTROLDIR);
+  control_dir = str_fmt("%s/%s", dir, EXTRACTCONTROLDIR);
 
   data_options = DPKG_TAR_EXTRACT | DPKG_TAR_PERMS;
   if (opt_verbose)
