@@ -1,7 +1,7 @@
 /*
  * libcompat - system compatibility library
  *
- * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,10 +22,16 @@
 #include <stdio.h>
 #include <gettext.h>
 
+#include "compat.h"
+
 #define _(str) gettext(str)
 
+#if !HAVE_DECL_SYS_ERRLIST
 extern const char *const sys_errlist[];
+#endif
+#if !HAVE_DECL_SYS_NERR
 extern const int sys_nerr;
+#endif
 
 const char *
 strerror(int e)
