@@ -4,7 +4,7 @@
  *
  * Copyright © 2000 Wichert Akkerman <wakkerma@debian.org>
  * Copyright © 2004 Scott James Remnant <scott@netsplit.com>
- * Copyright © 2006-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2006-2015 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,6 +32,9 @@
 #ifdef WITH_ZLIB
 #include <zlib.h>
 #endif
+#ifdef WITH_LIBZSTD
+#include <zstd.h>
+#endif
 #ifdef WITH_LIBLZMA
 #include <lzma.h>
 #endif
@@ -47,7 +50,7 @@
 #include <dpkg/buffer.h>
 #include <dpkg/command.h>
 #include <dpkg/compress.h>
-#if !defined(WITH_ZLIB) || !defined(WITH_LIBLZMA) || !defined(WITH_BZ2)
+#if !defined(WITH_ZLIB) || !defined(WITH_LIBZSTD) || !defined(WITH_LIBLZMA) || !defined(WITH_BZ2)
 #include <dpkg/subproc.h>
 
 static void DPKG_ATTR_SENTINEL
@@ -81,7 +84,7 @@ fd_fd_filter(int fd_in, int fd_out, const char *desc, const char *delenv[],
 
 		command_exec(&cmd);
 	}
-	subproc_wait_check(pid, desc, 0);
+	subproc_reap(pid, desc, 0);
 }
 #endif
 
@@ -142,7 +145,7 @@ fixup_gzip_params(struct compress_params *params)
 {
 	/* Normalize compression level. */
 	if (params->level == 0)
-		params->type = compressor_type_none;
+		params->type = COMPRESSOR_TYPE_NONE;
 }
 
 #ifdef WITH_ZLIB
@@ -189,13 +192,13 @@ compress_gzip(int fd_in, int fd_out, struct compress_params *params, const char
 	int z_errnum;
 	gzFile gzfile;
 
-	if (params->strategy == compressor_strategy_filtered)
+	if (params->strategy == COMPRESSOR_STRATEGY_FILTERED)
 		strategy = 'f';
-	else if (params->strategy == compressor_strategy_huffman)
+	else if (params->strategy == COMPRESSOR_STRATEGY_HUFFMAN)
 		strategy = 'h';
-	else if (params->strategy == compressor_strategy_rle)
+	else if (params->strategy == COMPRESSOR_STRATEGY_RLE)
 		strategy = 'R';
-	else if (params->strategy == compressor_strategy_fixed)
+	else if (params->strategy == COMPRESSOR_STRATEGY_FIXED)
 		strategy = 'F';
 	else
 		strategy = ' ';
@@ -503,7 +506,7 @@ filter_lzma(struct io_lzma *io, int fd_in, int fd_out)
 		ohshite(_("%s: lzma close error"), io->desc);
 }
 
-static void
+static void DPKG_ATTR_NORET
 filter_lzma_error(struct io_lzma *io, lzma_ret ret)
 {
 	ohshit(_("%s: lzma error: %s"), io->desc,
@@ -527,14 +530,32 @@ static void
 filter_xz_init(struct io_lzma *io, lzma_stream *s)
 {
 	uint32_t preset;
+	lzma_check check = LZMA_CHECK_CRC64;
+#ifdef HAVE_LZMA_MT
+	lzma_mt mt_options = {
+		.flags = 0,
+		.threads = sysconf(_SC_NPROCESSORS_ONLN),
+		.block_size = 0,
+		.timeout = 0,
+		.filters = NULL,
+		.check = check,
+	};
+#endif
 	lzma_ret ret;
 
 	io->status |= DPKG_STREAM_COMPRESS;
 
 	preset = io->params->level;
-	if (io->params->strategy == compressor_strategy_extreme)
+	if (io->params->strategy == COMPRESSOR_STRATEGY_EXTREME)
 		preset |= LZMA_PRESET_EXTREME;
-	ret = lzma_easy_encoder(s, preset, LZMA_CHECK_CRC32);
+
+#ifdef HAVE_LZMA_MT
+	mt_options.preset = preset;
+	ret = lzma_stream_encoder_mt(s, &mt_options);
+#else
+	ret = lzma_easy_encoder(s, preset, check);
+#endif
+
 	if (ret != LZMA_OK)
 		filter_lzma_error(io, ret);
 }
@@ -598,7 +619,7 @@ compress_xz(int fd_in, int fd_out, struct compress_params *params, const char *d
 	char combuf[6];
 	const char *strategy;
 
-	if (params->strategy == compressor_strategy_extreme)
+	if (params->strategy == COMPRESSOR_STRATEGY_EXTREME)
 		strategy = "-e";
 	else
 		strategy = NULL;
@@ -645,7 +666,7 @@ filter_lzma_init(struct io_lzma *io, lzma_stream *s)
 	io->status |= DPKG_STREAM_COMPRESS;
 
 	preset = io->params->level;
-	if (io->params->strategy == compressor_strategy_extreme)
+	if (io->params->strategy == COMPRESSOR_STRATEGY_EXTREME)
 		preset |= LZMA_PRESET_EXTREME;
 	if (lzma_lzma_preset(&options, preset))
 		filter_lzma_error(io, LZMA_OPTIONS_ERROR);
@@ -707,16 +728,98 @@ static const struct compressor compressor_lzma = {
 	.decompress = decompress_lzma,
 };
 
+/*
+ * Zstd compressor.
+ */
+
+#define ZSTD		"zstd"
+
+#ifdef WITH_LIBZSTD
+
+static void
+decompress_zstd(int fd_in, int fd_out, const char *desc)
+{
+	size_t const buf_in_size = ZSTD_DStreamInSize();
+	void*  const buf_in = m_malloc(buf_in_size);
+	size_t const buf_out_size = ZSTD_DStreamOutSize();
+	void*  const buf_out = m_malloc(buf_out_size);
+	size_t init_result, just_read, to_read;
+	ZSTD_DStream* const dstream = ZSTD_createDStream();
+	if (dstream == NULL) {
+		ohshit(_("ZSTD_createDStream error creating stream"));
+	}
+
+	init_result = ZSTD_initDStream(dstream);
+	if (ZSTD_isError(init_result)) {
+		ohshit(_("ZSTD_initDStream error : %s"), ZSTD_getErrorName(init_result));
+	}
+	to_read = init_result;
+	while ((just_read = fd_read(fd_in, buf_in, to_read))) {
+		ZSTD_inBuffer input = { buf_in, just_read, 0 };
+		while (input.pos < input.size) {
+			ssize_t actualwrite;
+			ZSTD_outBuffer output = { buf_out, buf_out_size, 0 };
+			to_read = ZSTD_decompressStream(dstream, &output , &input);
+			if (ZSTD_isError(to_read)) {
+				ohshit(_("ZSTD_decompressStream error : %s \n"),
+				       ZSTD_getErrorName(to_read));
+			}
+			actualwrite = fd_write(fd_out, output.dst, output.pos);
+			if (actualwrite != output.pos) {
+				const char *errmsg = strerror(errno);
+				ohshite(_("%s: internal zstd write error: '%s'"), desc, errmsg);
+			}
+			/* possible next frame */
+			if (to_read == 0) {
+				init_result = ZSTD_initDStream(dstream);
+				if (ZSTD_isError(init_result)) {
+					ohshit(_("ZSTD_initDStream error : %s"), ZSTD_getErrorName(init_result));
+				}
+				to_read = init_result;
+			}
+		}
+	}
+
+	ZSTD_freeDStream(dstream);
+	free(buf_in);
+	free(buf_out);
+	if (close(fd_out))
+		ohshite(_("%s: internal zstd write error"), desc);
+}
+
+#else
+static const char *env_zstd[] = {};
+
+static void
+decompress_zstd(int fd_in, int fd_out, const char *desc)
+{
+	fd_fd_filter(fd_in, fd_out, desc, env_zstd, ZSTD, "-dcq", NULL);
+}
+
+#endif
+
+static const struct compressor compressor_zstd = {
+	.name = "zstd",
+	.extension = ".zst",
+        /* zstd commands's default is 3 but the aim is to be closer to xz's
+         * default compression efficiency */
+	.default_level = 19,
+	.fixup_params = fixup_none_params,
+	.compress = NULL,
+	.decompress = decompress_zstd,
+};
+
 /*
  * Generic compressor filter.
  */
 
 static const struct compressor *compressor_array[] = {
-	[compressor_type_none] = &compressor_none,
-	[compressor_type_gzip] = &compressor_gzip,
-	[compressor_type_xz] = &compressor_xz,
-	[compressor_type_bzip2] = &compressor_bzip2,
-	[compressor_type_lzma] = &compressor_lzma,
+	[COMPRESSOR_TYPE_NONE] = &compressor_none,
+	[COMPRESSOR_TYPE_GZIP] = &compressor_gzip,
+	[COMPRESSOR_TYPE_XZ] = &compressor_xz,
+	[COMPRESSOR_TYPE_BZIP2] = &compressor_bzip2,
+	[COMPRESSOR_TYPE_LZMA] = &compressor_lzma,
+	[COMPRESSOR_TYPE_ZSTD] = &compressor_zstd,
 };
 
 static const struct compressor *
@@ -730,6 +833,12 @@ compressor(enum compressor_type type)
 	return compressor_array[type];
 }
 
+const char *
+compressor_get_name(enum compressor_type type)
+{
+	return compressor(type)->name;
+}
+
 const char *
 compressor_get_extension(enum compressor_type type)
 {
@@ -745,7 +854,7 @@ compressor_find_by_name(const char *name)
 		if (strcmp(compressor_array[i]->name, name) == 0)
 			return i;
 
-	return compressor_type_unknown;
+	return COMPRESSOR_TYPE_UNKNOWN;
 }
 
 enum compressor_type
@@ -757,53 +866,58 @@ compressor_find_by_extension(const char *extension)
 		if (strcmp(compressor_array[i]->extension, extension) == 0)
 			return i;
 
-	return compressor_type_unknown;
+	return COMPRESSOR_TYPE_UNKNOWN;
 }
 
 enum compressor_strategy
 compressor_get_strategy(const char *name)
 {
 	if (strcmp(name, "none") == 0)
-		return compressor_strategy_none;
+		return COMPRESSOR_STRATEGY_NONE;
 	if (strcmp(name, "filtered") == 0)
-		return compressor_strategy_filtered;
+		return COMPRESSOR_STRATEGY_FILTERED;
 	if (strcmp(name, "huffman") == 0)
-		return compressor_strategy_huffman;
+		return COMPRESSOR_STRATEGY_HUFFMAN;
 	if (strcmp(name, "rle") == 0)
-		return compressor_strategy_rle;
+		return COMPRESSOR_STRATEGY_RLE;
 	if (strcmp(name, "fixed") == 0)
-		return compressor_strategy_fixed;
+		return COMPRESSOR_STRATEGY_FIXED;
 	if (strcmp(name, "extreme") == 0)
-		return compressor_strategy_extreme;
+		return COMPRESSOR_STRATEGY_EXTREME;
 
-	return compressor_strategy_unknown;
-}
-
-bool
-compressor_check_params(struct compress_params *params, struct dpkg_error *err)
-{
-	if (params->strategy == compressor_strategy_none)
-		return true;
-
-	if (params->type == compressor_type_gzip &&
-	    (params->strategy == compressor_strategy_filtered ||
-	     params->strategy == compressor_strategy_huffman ||
-	     params->strategy == compressor_strategy_rle ||
-	     params->strategy == compressor_strategy_fixed))
-		return true;
-
-	if (params->type == compressor_type_xz &&
-	    params->strategy == compressor_strategy_extreme)
-		return true;
-
-	dpkg_put_error(err, _("unknown compression strategy"));
-	return false;
+	return COMPRESSOR_STRATEGY_UNKNOWN;
 }
 
 static void
 compressor_fixup_params(struct compress_params *params)
 {
 	compressor(params->type)->fixup_params(params);
+
+	if (params->level < 0)
+		params->level = compressor(params->type)->default_level;
+}
+
+bool
+compressor_check_params(struct compress_params *params, struct dpkg_error *err)
+{
+	compressor_fixup_params(params);
+
+	if (params->strategy == COMPRESSOR_STRATEGY_NONE)
+		return true;
+
+	if (params->type == COMPRESSOR_TYPE_GZIP &&
+	    (params->strategy == COMPRESSOR_STRATEGY_FILTERED ||
+	     params->strategy == COMPRESSOR_STRATEGY_HUFFMAN ||
+	     params->strategy == COMPRESSOR_STRATEGY_RLE ||
+	     params->strategy == COMPRESSOR_STRATEGY_FIXED))
+		return true;
+
+	if (params->type == COMPRESSOR_TYPE_XZ &&
+	    params->strategy == COMPRESSOR_STRATEGY_EXTREME)
+		return true;
+
+	dpkg_put_error(err, _("unknown compression strategy"));
+	return false;
 }
 
 void
@@ -818,6 +932,8 @@ decompress_filter(enum compressor_type type, int fd_in, int fd_out,
 	va_end(args);
 
 	compressor(type)->decompress(fd_in, fd_out, desc.buf);
+
+	varbuf_destroy(&desc);
 }
 
 void
@@ -831,10 +947,7 @@ compress_filter(struct compress_params *params, int fd_in, int fd_out,
 	varbuf_vprintf(&desc, desc_fmt, args);
 	va_end(args);
 
-	compressor_fixup_params(params);
-
-	if (params->level < 0)
-		params->level = compressor(params->type)->default_level;
-
 	compressor(params->type)->compress(fd_in, fd_out, params, desc.buf);
+
+	varbuf_destroy(&desc);
 }
