@@ -3,7 +3,8 @@
  * trigdeferred.h - parsing of triggers/Deferred
  *
  * Copyright © 2007 Canonical, Ltd.
- *   written by Ian Jackson <ian@chiark.greenend.org.uk>
+ *   written by Ian Jackson <ijackson@chiark.greenend.org.uk>
+ * Copyright © 2008-2014 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,21 +33,21 @@ DPKG_BEGIN_DECLS
  * @{
  */
 
-enum trigdef_updateflags {
-	tduf_nolockok		= DPKG_BIT(0),
-	tduf_write		= DPKG_BIT(1),
-	tduf_nolock		= tduf_nolockok | tduf_write,
-	/** Should not be set unless _write is. */
-	tduf_writeifempty	= DPKG_BIT(3),
-	tduf_writeifenoent	= DPKG_BIT(4),
+enum trigdef_update_flags {
+	TDUF_NO_LOCK_OK		= DPKG_BIT(0),
+	TDUF_WRITE		= DPKG_BIT(1),
+	TDUF_NO_LOCK		= TDUF_NO_LOCK_OK | TDUF_WRITE,
+	/** Should not be set unless TDUF_WRITE is. */
+	TDUF_WRITE_IF_EMPTY	= DPKG_BIT(3),
+	TDUF_WRITE_IF_ENOENT	= DPKG_BIT(4),
 };
 
 enum trigdef_update_status {
-	tdus_error_no_dir = -1,
-	tdus_error_empty_deferred = -2,
-	tdus_error_no_deferred = -3,
-	tdus_no_deferred = 1,
-	tdus_ok = 2,
+	TDUS_ERROR_NO_DIR = -1,
+	TDUS_ERROR_EMPTY_DEFERRED = -2,
+	TDUS_ERROR_NO_DEFERRED = -3,
+	TDUS_NO_DEFERRED = 1,
+	TDUS_OK = 2,
 };
 
 struct trigdefmeths {
@@ -57,7 +58,7 @@ struct trigdefmeths {
 
 void trigdef_set_methods(const struct trigdefmeths *methods);
 
-enum trigdef_update_status trigdef_update_start(enum trigdef_updateflags uf);
+enum trigdef_update_status trigdef_update_start(enum trigdef_update_flags uf);
 void trigdef_update_printf(const char *format, ...) DPKG_ATTR_PRINTF(1);
 int trigdef_parse(void);
 void trigdef_process_done(void);
