@@ -1,4 +1,5 @@
 # Copyright © 2008-2011 Raphaël Hertzog <hertzog@debian.org>
+# Copyright © 2008-2015 Guillem Jover <guillem@debian.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,12 +21,20 @@ use warnings;
 
 our $VERSION = '0.01';
 
-use parent qw(Dpkg::Source::Package);
+use POSIX qw(:errno_h);
+use List::Util qw(first);
+use Cwd;
+use File::Basename;
+use File::Temp qw(tempfile tempdir);
+use File::Path qw(make_path);
+use File::Spec;
+use File::Find;
+use File::Copy;
 
-use Dpkg;
 use Dpkg::Gettext;
 use Dpkg::ErrorHandling;
 use Dpkg::File;
+use Dpkg::Path qw(find_command);
 use Dpkg::Compression;
 use Dpkg::Source::Archive;
 use Dpkg::Source::Patch;
@@ -35,75 +44,63 @@ use Dpkg::Vendor qw(run_vendor_hook);
 use Dpkg::Control;
 use Dpkg::Changelog::Parse;
 
-use POSIX qw(:errno_h);
-use Cwd;
-use File::Basename;
-use File::Temp qw(tempfile tempdir);
-use File::Path;
-use File::Spec;
-use File::Find;
-use File::Copy;
+use parent qw(Dpkg::Source::Package);
 
 our $CURRENT_MINOR_VERSION = '0';
 
 sub init_options {
-    my ($self) = @_;
+    my $self = shift;
     $self->SUPER::init_options();
-    $self->{options}{include_removal} = 0
-        unless exists $self->{options}{include_removal};
-    $self->{options}{include_timestamp} = 0
-        unless exists $self->{options}{include_timestamp};
-    $self->{options}{include_binaries} = 0
-        unless exists $self->{options}{include_binaries};
-    $self->{options}{preparation} = 1
-        unless exists $self->{options}{preparation};
-    $self->{options}{skip_patches} = 0
-        unless exists $self->{options}{skip_patches};
-    $self->{options}{unapply_patches} = 'auto'
-        unless exists $self->{options}{unapply_patches};
-    $self->{options}{skip_debianization} = 0
-        unless exists $self->{options}{skip_debianization};
-    $self->{options}{create_empty_orig} = 0
-        unless exists $self->{options}{create_empty_orig};
-    $self->{options}{auto_commit} = 0
-        unless exists $self->{options}{auto_commit};
+    $self->{options}{include_removal} //= 0;
+    $self->{options}{include_timestamp} //= 0;
+    $self->{options}{include_binaries} //= 0;
+    $self->{options}{preparation} //= 1;
+    $self->{options}{skip_patches} //= 0;
+    $self->{options}{unapply_patches} //= 'auto';
+    $self->{options}{skip_debianization} //= 0;
+    $self->{options}{create_empty_orig} //= 0;
+    $self->{options}{auto_commit} //= 0;
+    $self->{options}{ignore_bad_version} //= 0;
 }
 
 sub parse_cmdline_option {
     my ($self, $opt) = @_;
-    if ($opt =~ /^--include-removal$/) {
+    if ($opt eq '--include-removal') {
         $self->{options}{include_removal} = 1;
         return 1;
-    } elsif ($opt =~ /^--include-timestamp$/) {
+    } elsif ($opt eq '--include-timestamp') {
         $self->{options}{include_timestamp} = 1;
         return 1;
-    } elsif ($opt =~ /^--include-binaries$/) {
+    } elsif ($opt eq '--include-binaries') {
         $self->{options}{include_binaries} = 1;
         return 1;
-    } elsif ($opt =~ /^--no-preparation$/) {
+    } elsif ($opt eq '--no-preparation') {
         $self->{options}{preparation} = 0;
         return 1;
-    } elsif ($opt =~ /^--skip-patches$/) {
+    } elsif ($opt eq '--skip-patches') {
         $self->{options}{skip_patches} = 1;
         return 1;
-    } elsif ($opt =~ /^--unapply-patches$/) {
+    } elsif ($opt eq '--unapply-patches') {
         $self->{options}{unapply_patches} = 'yes';
         return 1;
-    } elsif ($opt =~ /^--no-unapply-patches$/) {
+    } elsif ($opt eq '--no-unapply-patches') {
         $self->{options}{unapply_patches} = 'no';
         return 1;
-    } elsif ($opt =~ /^--skip-debianization$/) {
+    } elsif ($opt eq '--skip-debianization') {
         $self->{options}{skip_debianization} = 1;
         return 1;
-    } elsif ($opt =~ /^--create-empty-orig$/) {
+    } elsif ($opt eq '--create-empty-orig') {
         $self->{options}{create_empty_orig} = 1;
         return 1;
-    } elsif ($opt =~ /^--abort-on-upstream-changes$/) {
+    } elsif ($opt eq '--abort-on-upstream-changes') {
         $self->{options}{auto_commit} = 0;
         return 1;
-    } elsif ($opt =~ /^--auto-commit$/) {
+    } elsif ($opt eq '--auto-commit') {
         $self->{options}{auto_commit} = 1;
         return 1;
+    } elsif ($opt eq '--ignore-bad-version') {
+        $self->{options}{ignore_bad_version} = 1;
+        return 1;
     }
     return 0;
 }
@@ -117,14 +114,14 @@ sub do_extract {
     my $basename = $self->get_basename();
     my $basenamerev = $self->get_basename(1);
 
-    my ($tarfile, $debianfile, %origtar, %seen);
-    my ($tarsign, %origtarsign);
+    my ($tarfile, $debianfile, %addonfile, %seen);
+    my ($tarsign, %addonsign);
     my $re_ext = compression_get_file_extension_regex();
     foreach my $file ($self->get_files()) {
         my $uncompressed = $file;
         $uncompressed =~ s/\.$re_ext$/.*/;
         $uncompressed =~ s/\.$re_ext\.asc$/.*.asc/;
-        error(_g('duplicate files in %s source package: %s'), 'v2.0',
+        error(g_('duplicate files in %s source package: %s'), 'v2.0',
               $uncompressed) if $seen{$uncompressed};
         $seen{$uncompressed} = 1;
         if ($file =~ /^\Q$basename\E\.orig\.tar\.$re_ext$/) {
@@ -132,37 +129,37 @@ sub do_extract {
         } elsif ($file =~ /^\Q$basename\E\.orig\.tar\.$re_ext\.asc$/) {
             $tarsign = $file;
         } elsif ($file =~ /^\Q$basename\E\.orig-([[:alnum:]-]+)\.tar\.$re_ext$/) {
-            $origtar{$1} = $file;
+            $addonfile{$1} = $file;
         } elsif ($file =~ /^\Q$basename\E\.orig-([[:alnum:]-]+)\.tar\.$re_ext\.asc$/) {
-            $origtarsign{$1} = $file;
+            $addonsign{$1} = $file;
         } elsif ($file =~ /^\Q$basenamerev\E\.debian\.tar\.$re_ext$/) {
             $debianfile = $file;
         } else {
-            error(_g('unrecognized file for a %s source package: %s'),
+            error(g_('unrecognized file for a %s source package: %s'),
             'v2.0', $file);
         }
     }
 
     unless ($tarfile and $debianfile) {
-        error(_g('missing orig.tar or debian.tar file in v2.0 source package'));
+        error(g_('missing orig.tar or debian.tar file in v2.0 source package'));
     }
     if ($tarsign and $tarfile ne substr $tarsign, 0, -4) {
-        error(_g('mismatched orig.tar %s for signature %s in source package'),
+        error(g_('mismatched orig.tar %s for signature %s in source package'),
               $tarfile, $tarsign);
     }
-    foreach my $name (keys %origtarsign) {
-        error(_g('missing addon orig.tar for signature %s in source package'),
-              $origtarsign{$name})
-            if not exists $origtar{$name};
-        error(_g('mismatched addon orig.tar %s for signature %s in source package'),
-              $origtar{$name}, $origtarsign{$name})
-            if $origtar{$name} ne substr $origtarsign{$name}, 0, -4;
+    foreach my $name (keys %addonsign) {
+        error(g_('missing addon orig.tar for signature %s in source package'),
+              $addonsign{$name})
+            if not exists $addonfile{$name};
+        error(g_('mismatched addon orig.tar %s for signature %s in source package'),
+              $addonfile{$name}, $addonsign{$name})
+            if $addonfile{$name} ne substr $addonsign{$name}, 0, -4;
     }
 
     erasedir($newdirectory);
 
     # Extract main tarball
-    info(_g('unpacking %s'), $tarfile);
+    info(g_('unpacking %s'), $tarfile);
     my $tar = Dpkg::Source::Archive->new(filename => "$dscdir$tarfile");
     $tar->extract($newdirectory, no_fixperms => 1,
                   options => [ '--anchored', '--no-wildcards-match-slash',
@@ -172,11 +169,16 @@ sub do_extract {
     # that would be blindly followed when applying the patches
 
     # Extract additional orig tarballs
-    foreach my $subdir (keys %origtar) {
-        my $file = $origtar{$subdir};
-        info(_g('unpacking %s'), $file);
+    foreach my $subdir (sort keys %addonfile) {
+        my $file = $addonfile{$subdir};
+        info(g_('unpacking %s'), $file);
+
+        # If the pathname is an empty directory, just silently remove it, as
+        # it might be part of a git repository, as a submodule for example.
+        rmdir "$newdirectory/$subdir";
         if (-e "$newdirectory/$subdir") {
-            warning(_g("required removal of `%s' installed by original tarball"), $subdir);
+            warning(g_("required removal of '%s' installed by original tarball"),
+                    $subdir);
             erasedir("$newdirectory/$subdir");
         }
         $tar = Dpkg::Source::Archive->new(filename => "$dscdir$file");
@@ -187,24 +189,12 @@ sub do_extract {
     return if $self->{options}{skip_debianization};
 
     # Extract debian tarball after removing the debian directory
-    info(_g('unpacking %s'), $debianfile);
+    info(g_('unpacking %s'), $debianfile);
     erasedir("$newdirectory/debian");
-    # Exclude existing symlinks from extraction of debian.tar.gz as we
-    # don't want to overwrite something outside of $newdirectory due to a
-    # symlink
-    my @exclude_symlinks;
-    my $wanted = sub {
-        return if not -l $_;
-        my $fn = File::Spec->abs2rel($_, $newdirectory);
-        push @exclude_symlinks, '--exclude', $fn;
-    };
-    find({ wanted => $wanted, no_chdir => 1 }, $newdirectory);
     $tar = Dpkg::Source::Archive->new(filename => "$dscdir$debianfile");
-    $tar->extract($newdirectory, in_place => 1,
-                  options => [ '--anchored', '--no-wildcards',
-                  @exclude_symlinks ]);
+    $tar->extract($newdirectory, in_place => 1);
 
-    # Apply patches (in a separate method as it might be overriden)
+    # Apply patches (in a separate method as it might be overridden)
     $self->apply_patches($newdirectory, usage => 'unpack')
         unless $self->{options}{skip_patches};
 }
@@ -220,7 +210,7 @@ sub get_patches {
     my $pd = "$dir/debian/patches";
     my $auto_patch = $self->get_autopatch_name();
     if (-d $pd) {
-        opendir(my $dir_dh, $pd) or syserr(_g('cannot opendir %s'), $pd);
+        opendir(my $dir_dh, $pd) or syserr(g_('cannot opendir %s'), $pd);
         foreach my $patch (sort readdir($dir_dh)) {
             # patches match same rules as run-parts
             next unless $patch =~ /^[\w-]+$/ and -f "$pd/$patch";
@@ -239,12 +229,12 @@ sub apply_patches {
     return unless scalar(@patches);
     my $applied = File::Spec->catfile($dir, 'debian', 'patches', '.dpkg-source-applied');
     open(my $applied_fh, '>', $applied)
-        or syserr(_g('cannot write %s'), $applied);
+        or syserr(g_('cannot write %s'), $applied);
     print { $applied_fh } "# During $opts{usage}\n";
     my $timestamp = fs_time($applied);
     foreach my $patch ($self->get_patches($dir, %opts)) {
         my $path = File::Spec->catfile($dir, 'debian', 'patches', $patch);
-        info(_g('applying %s'), $patch) unless $opts{skip_auto};
+        info(g_('applying %s'), $patch) unless $opts{skip_auto};
         my $patch_obj = Dpkg::Source::Patch->new(filename => $path);
         $patch_obj->apply($dir, force_timestamp => 1,
                           timestamp => $timestamp,
@@ -262,7 +252,7 @@ sub unapply_patches {
     my $timestamp = fs_time($applied);
     foreach my $patch (@patches) {
         my $path = File::Spec->catfile($dir, 'debian', 'patches', $patch);
-        info(_g('unapplying %s'), $patch) unless $opts{quiet};
+        info(g_('unapplying %s'), $patch) unless $opts{quiet};
         my $patch_obj = Dpkg::Source::Patch->new(filename => $path);
         $patch_obj->apply($dir, force_timestamp => 1, verbose => 0,
                           timestamp => $timestamp,
@@ -272,7 +262,7 @@ sub unapply_patches {
 }
 
 sub upstream_tarball_template {
-    my ($self) = @_;
+    my $self = shift;
     my $ext = '{' . join(',',
         sort map {
             compression_get_property($_, 'file_ext')
@@ -285,7 +275,7 @@ sub can_build {
     return 1 if $self->find_original_tarballs(include_supplementary => 0);
     return 1 if $self->{options}{create_empty_orig} and
                 $self->find_original_tarballs(include_main => 0);
-    return (0, sprintf(_g('no upstream tarball found at %s'),
+    return (0, sprintf(g_('no upstream tarball found at %s'),
                        $self->upstream_tarball_template()));
 }
 
@@ -300,7 +290,7 @@ sub after_build {
     my $reason = '';
     if (-e $applied) {
         open(my $applied_fh, '<', $applied)
-            or syserr(_g('cannot read %s'), $applied);
+            or syserr(g_('cannot read %s'), $applied);
         $reason = <$applied_fh>;
         close($applied_fh);
     }
@@ -328,7 +318,8 @@ sub prepare_build {
         # No main orig.tar, create a dummy one
         my $filename = $self->get_basename() . '.orig.tar.' .
                        $self->{options}{comp_ext};
-        my $tar = Dpkg::Source::Archive->new(filename => $filename);
+        my $tar = Dpkg::Source::Archive->new(filename => $filename,
+                                             compression_level => $self->{options}{comp_level});
         $tar->create();
         $tar->finish();
     }
@@ -338,7 +329,7 @@ sub check_patches_applied {
     my ($self, $dir) = @_;
     my $applied = File::Spec->catfile($dir, 'debian', 'patches', '.dpkg-source-applied');
     unless (-e $applied) {
-        info(_g('patches are not applied, applying them now'));
+        info(g_('patches are not applied, applying them now'));
         $self->apply_patches($dir, usage => 'preparation');
     }
 }
@@ -350,30 +341,30 @@ sub generate_patch {
     $basedirname =~ s/_/-/;
 
     # Identify original tarballs
-    my ($tarfile, %origtar);
+    my ($tarfile, %addonfile);
     my $comp_ext_regex = compression_get_file_extension_regex();
     my @origtarballs;
-    foreach (sort $self->find_original_tarballs()) {
-        if (/\.orig\.tar\.$comp_ext_regex$/) {
+    foreach my $file (sort $self->find_original_tarballs()) {
+        if ($file =~ /\.orig\.tar\.$comp_ext_regex$/) {
             if (defined($tarfile)) {
-                error(_g('several orig.tar files found (%s and %s) but only ' .
-                         'one is allowed'), $tarfile, $_);
+                error(g_('several orig.tar files found (%s and %s) but only ' .
+                         'one is allowed'), $tarfile, $file);
             }
-            $tarfile = $_;
-            push @origtarballs, $_;
-            $self->add_file($_);
-        } elsif (/\.orig-([[:alnum:]-]+)\.tar\.$comp_ext_regex$/) {
-            $origtar{$1} = $_;
-            push @origtarballs, $_;
-            $self->add_file($_);
+            $tarfile = $file;
+            push @origtarballs, $file;
+            $self->add_file($file);
+        } elsif ($file =~ /\.orig-([[:alnum:]-]+)\.tar\.$comp_ext_regex$/) {
+            $addonfile{$1} = $file;
+            push @origtarballs, $file;
+            $self->add_file($file);
         }
     }
 
-    error(_g('no upstream tarball found at %s'),
+    error(g_('no upstream tarball found at %s'),
           $self->upstream_tarball_template()) unless $tarfile;
 
     if ($opts{usage} eq 'build') {
-        info(_g('building %s using existing %s'),
+        info(g_('building %s using existing %s'),
              $self->{fields}{'Source'}, "@origtarballs");
     }
 
@@ -386,8 +377,8 @@ sub generate_patch {
     $tar->extract($tmp);
 
     # Extract additional orig tarballs
-    foreach my $subdir (keys %origtar) {
-        my $file = $origtar{$subdir};
+    foreach my $subdir (keys %addonfile) {
+        my $file = $addonfile{$subdir};
         $tar = Dpkg::Source::Archive->new(filename => $file);
         $tar->extract("$tmp/$subdir");
     }
@@ -395,7 +386,7 @@ sub generate_patch {
     # Copy over the debian directory
     erasedir("$tmp/debian");
     system('cp', '-a', '--', "$dir/debian", "$tmp/");
-    subprocerr(_g('copy of the debian directory')) if $?;
+    subprocerr(g_('copy of the debian directory')) if $?;
 
     # Apply all patches except the last automatic one
     $opts{skip_auto} //= 0;
@@ -403,7 +394,7 @@ sub generate_patch {
 
     # Create a patch
     my ($difffh, $tmpdiff) = tempfile($self->get_basename(1) . '.diff.XXXXXX',
-                                      DIR => File::Spec->tmpdir(), UNLINK => 0);
+                                      TMPDIR => 1, UNLINK => 0);
     push_exit_handler(sub { unlink($tmpdiff) });
     my $diff = Dpkg::Source::Patch->new(filename => $tmpdiff,
                                         compression => 'none');
@@ -420,10 +411,10 @@ sub generate_patch {
             %{$self->{diff_options}},
             handle_binary_func => $opts{handle_binary},
             order_from => $opts{order_from});
-    error(_g('unrepresentable changes to source')) if not $diff->finish();
+    error(g_('unrepresentable changes to source')) if not $diff->finish();
 
     if (-s $tmpdiff) {
-        info(_g('local changes detected, the modified files are:'));
+        info(g_('local changes detected, the modified files are:'));
         my $analysis = $diff->analyze($dir, verbose => 0);
         foreach my $fn (sort keys %{$analysis->{filepatched}}) {
             print " $fn\n";
@@ -442,7 +433,7 @@ sub do_build {
     my ($self, $dir) = @_;
     my @argv = @{$self->{options}{ARGV}};
     if (scalar(@argv)) {
-        usageerr(_g("-b takes only one parameter with format `%s'"),
+        usageerr(g_("-b takes only one parameter with format '%s'"),
                  $self->{fields}{'Format'});
     }
     $self->prepare_build($dir);
@@ -457,21 +448,17 @@ sub do_build {
     my $binaryfiles = Dpkg::Source::Package::V2::BinaryFiles->new($dir);
     my $unwanted_binaries = 0;
     my $check_binary = sub {
-        if (-f $_ and is_binary($_)) {
+        if (-f and is_binary($_)) {
             my $fn = File::Spec->abs2rel($_, $dir);
             $binaryfiles->new_binary_found($fn);
             unless ($include_binaries or $binaryfiles->binary_is_allowed($fn)) {
-                errormsg(_g('unwanted binary file: %s'), $fn);
+                errormsg(g_('unwanted binary file: %s'), $fn);
                 $unwanted_binaries++;
             }
         }
     };
     my $tar_ignore_glob = '{' . join(',',
-        map {
-            my $copy = $_;
-            $copy =~ s/,/\\,/g;
-            $copy;
-        } @{$self->{options}{tar_ignore}}) . '}';
+        map { s/,/\\,/rg } @{$self->{options}{tar_ignore}}) . '}';
     my $filter_ignore = sub {
         # Filter out files that are not going to be included in the debian
         # tarball due to ignores.
@@ -479,13 +466,13 @@ sub do_build {
         my $reldir = File::Spec->abs2rel($File::Find::dir, $dir);
         my $cwd = getcwd();
         # Apply the pattern both from the top dir and from the inspected dir
-        chdir($dir) or syserr(_g("unable to chdir to `%s'"), $dir);
+        chdir $dir or syserr(g_("unable to chdir to '%s'"), $dir);
         $exclude{$_} = 1 foreach glob($tar_ignore_glob);
-        chdir($cwd) or syserr(_g("unable to chdir to `%s'"), $cwd);
+        chdir $cwd or syserr(g_("unable to chdir to '%s'"), $cwd);
         chdir($File::Find::dir)
-            or syserr(_g("unable to chdir to `%s'"), $File::Find::dir);
+            or syserr(g_("unable to chdir to '%s'"), $File::Find::dir);
         $exclude{$_} = 1 foreach glob($tar_ignore_glob);
-        chdir($cwd) or syserr(_g("unable to chdir to `%s'"), $cwd);
+        chdir $cwd or syserr(g_("unable to chdir to '%s'"), $cwd);
         my @result;
         foreach my $fn (@_) {
             unless (exists $exclude{$fn} or exists $exclude{"$reldir/$fn"}) {
@@ -505,15 +492,16 @@ sub do_build {
 
     # Handle modified binary files detected by the auto-patch generation
     my $handle_binary = sub {
-        my ($self, $old, $new) = @_;
-        my $relfn = File::Spec->abs2rel($new, $dir);
-        $binaryfiles->new_binary_found($relfn);
-        unless ($include_binaries or $binaryfiles->binary_is_allowed($relfn)) {
-            errormsg(_g('cannot represent change to %s: %s'), $relfn,
-                     _g('binary file contents changed'));
-            errormsg(_g('add %s in debian/source/include-binaries if you want ' .
+        my ($self, $old, $new, %opts) = @_;
+
+        my $file = $opts{filename};
+        $binaryfiles->new_binary_found($file);
+        unless ($include_binaries or $binaryfiles->binary_is_allowed($file)) {
+            errormsg(g_('cannot represent change to %s: %s'), $file,
+                     g_('binary file contents changed'));
+            errormsg(g_('add %s in debian/source/include-binaries if you want ' .
                         'to store the modified binary in the debian tarball'),
-                     $relfn);
+                     $file);
             $self->register_error();
         }
     };
@@ -527,9 +515,9 @@ sub do_build {
                                         skip_auto => $self->{options}{auto_commit},
                                         usage => 'build');
     unless (-z $tmpdiff or $self->{options}{auto_commit}) {
-        info(_g('you can integrate the local changes with %s'),
+        info(g_('you can integrate the local changes with %s'),
              'dpkg-source --commit');
-        error(_g('aborting due to unexpected upstream changes, see %s'),
+        error(g_('aborting due to unexpected upstream changes, see %s'),
               $tmpdiff);
     }
     push_exit_handler(sub { unlink($tmpdiff) });
@@ -537,20 +525,21 @@ sub do_build {
 
     # Install the diff as the new autopatch
     if ($self->{options}{auto_commit}) {
-        mkpath(File::Spec->catdir($dir, 'debian', 'patches'));
+        make_path(File::Spec->catdir($dir, 'debian', 'patches'));
         $autopatch = $self->register_patch($dir, $tmpdiff,
                                            $self->get_autopatch_name());
-        info(_g('local changes have been recorded in a new patch: %s'),
+        info(g_('local changes have been recorded in a new patch: %s'),
              $autopatch) if -e $autopatch;
         rmdir(File::Spec->catdir($dir, 'debian', 'patches')); # No check on purpose
     }
-    unlink($tmpdiff) or syserr(_g('cannot remove %s'), $tmpdiff);
+    unlink($tmpdiff) or syserr(g_('cannot remove %s'), $tmpdiff);
     pop_exit_handler();
 
     # Create the debian.tar
     my $debianfile = "$basenamerev.debian.tar." . $self->{options}{comp_ext};
-    info(_g('building %s in %s'), $sourcepackage, $debianfile);
-    my $tar = Dpkg::Source::Archive->new(filename => $debianfile);
+    info(g_('building %s in %s'), $sourcepackage, $debianfile);
+    my $tar = Dpkg::Source::Archive->new(filename => $debianfile,
+                                         compression_level => $self->{options}{comp_level});
     $tar->create(options => \@tar_ignore, chdir => $dir);
     $tar->add_directory('debian');
     foreach my $binary ($binaryfiles->get_seen_binaries()) {
@@ -569,7 +558,7 @@ sub get_patch_header {
     }
     my $text;
     if (-f $ph) {
-        open(my $ph_fh, '<', $ph) or syserr(_g('cannot read %s'), $ph);
+        open(my $ph_fh, '<', $ph) or syserr(g_('cannot read %s'), $ph);
         $text = file_slurp($ph_fh);
         close($ph_fh);
         return $text;
@@ -596,7 +585,7 @@ are templates for supplementary fields that you might want to add:
 
 Origin: <vendor|upstream|other>, <url of original patch>
 Bug: <url in upstream bugtracker>
-Bug-Debian: http://bugs.debian.org/<bugnumber>
+Bug-Debian: https://bugs.debian.org/<bugnumber>
 Bug-Ubuntu: https://launchpad.net/bugs/<bugnumber>
 Forwarded: <no|not-needed|url proving that it has been forwarded>
 Reviewed-By: <name and email of someone who approved the patch>
@@ -609,16 +598,16 @@ sub register_patch {
     my $patch = File::Spec->catfile($dir, 'debian', 'patches', $patch_name);
     if (-s $patch_file) {
         copy($patch_file, $patch)
-            or syserr(_g('failed to copy %s to %s'), $patch_file, $patch);
+            or syserr(g_('failed to copy %s to %s'), $patch_file, $patch);
         chmod(0666 & ~ umask(), $patch)
-            or syserr(_g("unable to change permission of `%s'"), $patch);
+            or syserr(g_("unable to change permission of '%s'"), $patch);
         my $applied = File::Spec->catfile($dir, 'debian', 'patches', '.dpkg-source-applied');
         open(my $applied_fh, '>>', $applied)
-            or syserr(_g('cannot write %s'), $applied);
+            or syserr(g_('cannot write %s'), $applied);
         print { $applied_fh } "$patch\n";
-        close($applied_fh) or syserr(_g('cannot close %s'), $applied);
+        close($applied_fh) or syserr(g_('cannot close %s'), $applied);
     } elsif (-e $patch) {
-        unlink($patch) or syserr(_g('cannot remove %s'), $patch);
+        unlink($patch) or syserr(g_('cannot remove %s'), $patch);
     }
     return $patch;
 }
@@ -631,7 +620,7 @@ sub _is_bad_patch_name {
 
     my $patch = File::Spec->catfile($dir, 'debian', 'patches', $patch_name);
     if (-e $patch) {
-        warning(_g('cannot register changes in %s, this patch already exists'),
+        warning(g_('cannot register changes in %s, this patch already exists'),
                 $patch);
         return 1;
     }
@@ -648,12 +637,12 @@ sub do_commit {
     if ($tmpdiff and not -e $tmpdiff) {
         $tmpdiff = File::Spec->catfile($dir, $tmpdiff)
             unless File::Spec->file_name_is_absolute($tmpdiff);
-        error(_g("patch file '%s' doesn't exist"), $tmpdiff) if not -e $tmpdiff;
+        error(g_("patch file '%s' doesn't exist"), $tmpdiff) if not -e $tmpdiff;
     }
 
     my $binaryfiles = Dpkg::Source::Package::V2::BinaryFiles->new($dir);
     my $handle_binary = sub {
-        my ($self, $old, $new) = @_;
+        my ($self, $old, $new, %opts) = @_;
         my $fn = File::Spec->abs2rel($new, $dir);
         $binaryfiles->new_binary_found($fn);
     };
@@ -665,24 +654,31 @@ sub do_commit {
     }
     push_exit_handler(sub { unlink($tmpdiff) });
     unless (-s $tmpdiff) {
-        unlink($tmpdiff) or syserr(_g('cannot remove %s'), $tmpdiff);
-        info(_g('there are no local changes to record'));
+        unlink($tmpdiff) or syserr(g_('cannot remove %s'), $tmpdiff);
+        info(g_('there are no local changes to record'));
         return;
     }
     while (_is_bad_patch_name($dir, $patch_name)) {
         # Ask the patch name interactively
-        print _g('Enter the desired patch name: ');
-        chomp($patch_name = <STDIN>);
+        print g_('Enter the desired patch name: ');
+        $patch_name = <STDIN>;
+        next unless defined $patch_name;
+        chomp $patch_name;
         $patch_name =~ s/\s+/-/g;
         $patch_name =~ s/\///g;
     }
-    mkpath(File::Spec->catdir($dir, 'debian', 'patches'));
+    make_path(File::Spec->catdir($dir, 'debian', 'patches'));
     my $patch = $self->register_patch($dir, $tmpdiff, $patch_name);
-    system('sensible-editor', $patch);
-    subprocerr('sensible-editor') if $?;
-    unlink($tmpdiff) or syserr(_g('cannot remove %s'), $tmpdiff);
+    my @editors = ('sensible-editor', $ENV{VISUAL}, $ENV{EDITOR}, 'vi');
+    my $editor = first { find_command($_) } @editors;
+    if (not $editor) {
+        error(g_('cannot find an editor'));
+    }
+    system($editor, $patch);
+    subprocerr($editor) if $?;
+    unlink($tmpdiff) or syserr(g_('cannot remove %s'), $tmpdiff);
     pop_exit_handler();
-    info(_g('local changes have been recorded in a new patch: %s'), $patch);
+    info(g_('local changes have been recorded in a new patch: %s'), $patch);
 }
 
 package Dpkg::Source::Package::V2::BinaryFiles;
@@ -690,7 +686,8 @@ package Dpkg::Source::Package::V2::BinaryFiles;
 use Dpkg::ErrorHandling;
 use Dpkg::Gettext;
 
-use File::Path;
+use File::Path qw(make_path);
+use File::Spec;
 
 sub new {
     my ($this, $dir) = @_;
@@ -715,14 +712,16 @@ sub new_binary_found {
 }
 
 sub load_allowed_binaries {
-    my ($self) = @_;
+    my $self = shift;
     my $incbin_file = $self->{include_binaries_path};
     if (-f $incbin_file) {
         open(my $incbin_fh, '<', $incbin_file)
-            or syserr(_g('cannot read %s'), $incbin_file);
-        while (defined($_ = <$incbin_fh>)) {
-            chomp; s/^\s*//; s/\s*$//;
-            next if /^#/ or /^$/;
+            or syserr(g_('cannot read %s'), $incbin_file);
+        while (<$incbin_fh>) {
+            chomp;
+            s/^\s*//;
+            s/\s*$//;
+            next if /^#/ or length == 0;
             $self->{allowed_binaries}{$_} = 1;
         }
         close($incbin_fh);
@@ -736,30 +735,30 @@ sub binary_is_allowed {
 }
 
 sub update_debian_source_include_binaries {
-    my ($self) = @_;
+    my $self = shift;
 
     my @unknown_binaries = $self->get_unknown_binaries();
     return unless scalar(@unknown_binaries);
 
     my $incbin_file = $self->{include_binaries_path};
-    mkpath(File::Spec->catdir($self->{dir}, 'debian', 'source'));
+    make_path(File::Spec->catdir($self->{dir}, 'debian', 'source'));
     open(my $incbin_fh, '>>', $incbin_file)
-        or syserr(_g('cannot write %s'), $incbin_file);
+        or syserr(g_('cannot write %s'), $incbin_file);
     foreach my $binary (@unknown_binaries) {
         print { $incbin_fh } "$binary\n";
-        info(_g('adding %s to %s'), $binary, 'debian/source/include-binaries');
+        info(g_('adding %s to %s'), $binary, 'debian/source/include-binaries');
         $self->{allowed_binaries}{$binary} = 1;
     }
     close($incbin_fh);
 }
 
 sub get_unknown_binaries {
-    my ($self) = @_;
+    my $self = shift;
     return grep { not $self->binary_is_allowed($_) } $self->get_seen_binaries();
 }
 
 sub get_seen_binaries {
-    my ($self) = @_;
+    my $self = shift;
     my @seen = sort keys %{$self->{seen_binaries}};
     return @seen;
 }
