@@ -1,4 +1,4 @@
-# Copyright © 2013 Guillem Jover <guillem@debian.org>
+# Copyright © 2013, 2015 Guillem Jover <guillem@debian.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,14 +19,22 @@ use strict;
 use warnings;
 
 our $VERSION = '0.01';
+our @EXPORT_OK = qw(
+    any
+    none
+);
+our %EXPORT_TAGS = (
+    list => [ qw(any none) ],
+);
 
 use Exporter qw(import);
-our @EXPORT_OK = qw(any none);
-our %EXPORT_TAGS = (list => [ qw(any none) ]);
 
 # XXX: Ideally we would use List::MoreUtils, but that's not a core module,
 # so to avoid the additional dependency we'll make do with the following
 # trivial reimplementations.
+#
+# These got added to List::Util 1.33, which got merged into perl 5.20.0,
+# once that is in Debian oldstable we can switch to that core module.
 
 sub any(&@) {
     my $code = shift;
