@@ -19,15 +19,15 @@
 use warnings;
 use strict;
 
+use Scalar::Util qw(blessed);
+use Getopt::Long qw(:config posix_default bundling no_ignorecase);
+
 use Dpkg ();
 use Dpkg::Changelog::Debian;
 use Dpkg::ErrorHandling;
 use Dpkg::Gettext;
 use Dpkg::Version;
 
-use Getopt::Long qw(:config posix_default bundling no_ignorecase);
-use Scalar::Util qw(blessed);
-
 textdomain('dpkg-dev');
 
 sub merge_entries($$$);
@@ -51,16 +51,16 @@ BEGIN {
 }
 
 sub version {
-    printf _g("Debian %s version %s.\n"), $Dpkg::PROGNAME, $Dpkg::PROGVERSION;
+    printf g_("Debian %s version %s.\n"), $Dpkg::PROGNAME, $Dpkg::PROGVERSION;
 
-    printf "\n" . _g(
+    printf "\n" . g_(
 'This is free software; see the GNU General Public License version 2 or
 later for copying conditions. There is NO warranty.
 ');
 }
 
 sub usage {
-    printf(_g(
+    printf g_(
 "Usage: %s [<option>...] <old> <new-a> <new-b> [<out>]
 
 Options:
@@ -68,7 +68,7 @@ Options:
                            after the last '~' in the version.
   -?, --help               show this help message.
       --version            show the version.
-"), $Dpkg::PROGNAME);
+"), $Dpkg::PROGNAME;
 }
 
 my $merge_prereleases;
@@ -87,11 +87,11 @@ my @options_spec = (
 my ($old, $new_a, $new_b, $out_file) = @ARGV;
 unless (defined $old and defined $new_a and defined $new_b)
 {
-    usageerr(_g('needs at least three arguments'));
+    usageerr(g_('needs at least three arguments'));
 }
 unless (-e $old and -e $new_a and -e $new_b)
 {
-    usageerr(_g('file arguments need to exist'));
+    usageerr(g_('file arguments need to exist'));
 }
 
 my ($cho, $cha, $chb);
@@ -110,7 +110,8 @@ my @result; # Lines to output
 my $exitcode = 0; # 1 if conflict encountered
 
 unless (merge_block($cho, $cha, $chb, sub {
-			my $tail = $_[0]->get_unparsed_tail();
+			my $changes = shift;
+			my $tail = $changes->get_unparsed_tail();
 			chomp $tail if defined $tail;
 			return $tail;
 		    }))
@@ -134,9 +135,9 @@ while (1) {
 
 if (defined($out_file) and $out_file ne '-') {
     open(my $out_fh, '>', $out_file)
-        or syserr(_g('cannot write %s'), $out_file);
+        or syserr(g_('cannot write %s'), $out_file);
     print { $out_fh } ((blessed $_) ? "$_" : "$_\n") foreach @result;
-    close($out_fh) or syserr(_g('cannot write %s'), $out_file);
+    close($out_fh) or syserr(g_('cannot write %s'), $out_file);
 } else {
     print ((blessed $_) ? "$_" : "$_\n") foreach @result;
 }
@@ -214,8 +215,9 @@ sub merge_entries($$$) {
 		       $a->get_part('changes'), $b->get_part('changes'),
 		       {
 			   CONFLICT => sub {
+				my ($ca, $cb) = @_;
 				$exitcode = 1;
-				return get_conflict_block($_[0], $_[1]);
+				return get_conflict_block($ca, $cb);
 			   }
 		       });
     unshift @result, @merged;
