@@ -5,7 +5,7 @@
 # Copyright © 1996 Ian Jackson
 # Copyright © 2000 Wichert Akkerman
 # Copyright © 2006 Frank Lichtenheld
-# Copyright © 2006-2010,2012-2013 Guillem Jover <guillem@debian.org>
+# Copyright © 2006-2010,2012-2015 Guillem Jover <guillem@debian.org>
 # Copyright © 2007 Raphaël Hertzog
 #
 # This program is free software; you can redistribute it and/or modify
@@ -38,6 +38,7 @@ use Dpkg::Version;
 use Dpkg::Shlibs qw(find_library get_library_paths);
 use Dpkg::Shlibs::Objdump;
 use Dpkg::Shlibs::SymbolFile;
+use Dpkg::Substvars;
 use Dpkg::Arch qw(get_host_arch);
 use Dpkg::Deps;
 use Dpkg::Control::Info;
@@ -63,9 +64,10 @@ my $shlibslocal = 'debian/shlibs.local';
 my $packagetype = 'deb';
 my $dependencyfield = 'Depends';
 my $varlistfile = 'debian/substvars';
+my $varlistfilenew;
 my $varnameprefix = 'shlibs';
 my $ignore_missing_info = 0;
-my $warnings = 3;
+my $warnings = WARN_SYM_NOT_FOUND | WARN_DEP_AVOIDABLE;
 my $debug = 0;
 my @exclude = ();
 my @pkg_dir_to_search = ();
@@ -102,13 +104,13 @@ foreach (@ARGV) {
     } elsif (m/^--admindir=(.*)$/) {
 	$admindir = $1;
 	if (not -d $admindir) {
-	    error(_g("administrative directory '%s' does not exist"), $admindir);
+	    error(g_("administrative directory '%s' does not exist"), $admindir);
 	}
 	$ENV{DPKG_ADMINDIR} = $admindir;
     } elsif (m/^-d(.*)$/) {
 	$dependencyfield = field_capitalize($1);
 	if (not defined $depstrength{$dependencyfield}) {
-	    warning(_g("unrecognized dependency field '%s'"), $dependencyfield);
+	    warning(g_("unrecognized dependency field '%s'"), $dependencyfield);
 	}
     } elsif (m/^-e(.*)$/) {
 	if (exists $exec{$1}) {
@@ -130,7 +132,7 @@ foreach (@ARGV) {
     } elsif (m/^-x(.*)$/) {
 	push @exclude, $1;
     } elsif (m/^-/) {
-	usageerr(_g("unknown option \`%s'"), $_);
+	usageerr(g_("unknown option '%s'"), $_);
     } else {
 	if (exists $exec{$_}) {
 	    # Affect the binary to the most important field
@@ -142,13 +144,13 @@ foreach (@ARGV) {
 	}
     }
 }
-usageerr(_g('need at least one executable')) unless scalar keys %exec;
+usageerr(g_('need at least one executable')) unless scalar keys %exec;
 
 my $control = Dpkg::Control::Info->new();
 my $fields = $control->get_source();
 my $bd_value = deps_concat($fields->{'Build-Depends'}, $fields->{'Build-Depends-Arch'});
 my $build_deps = deps_parse($bd_value, build_dep => 1, reduce_restrictions => 1);
-error(_g('error occurred while parsing %s'), 'Build-Depends/Build-Depends-Arch')
+error(g_('error occurred while parsing %s'), 'Build-Depends/Build-Depends-Arch')
     unless defined $build_deps;
 
 my %dependencies;
@@ -185,7 +187,7 @@ foreach my $file (keys %exec) {
 	unless (defined $lib) {
 	    $soname_notfound{$soname} = 1;
 	    $global_soname_notfound{$soname} = 1;
-	    my $msg = _g("couldn't find library %s needed by %s (ELF " .
+	    my $msg = g_("couldn't find library %s needed by %s (ELF " .
 			 "format: '%s'; RPATH: '%s')");
 	    if (scalar(split_soname($soname))) {
 		errormsg($msg, $soname, $file, $obj->{format}, join(':', @{$obj->{RPATH}}));
@@ -255,13 +257,9 @@ foreach my $file (keys %exec) {
                 # package and we really need it)
 		my $dep = $symfile->get_dependency($soname);
 		my $minver = $symfile->get_smallest_version($soname) || '';
-		foreach my $subdep (split /\s*,\s*/, $dep) {
-		    if (not exists $dependencies{$cur_field}{$subdep}) {
-			$dependencies{$cur_field}{$subdep} = Dpkg::Version->new($minver);
-                        print " Initialize dependency ($subdep) with minimal " .
-                              "version ($minver)\n" if $debug > 1;
-		    }
-		}
+		update_dependency_version($dep, $minver);
+		print " Minimal version of ($dep) initialized with ($minver)\n"
+		    if $debug > 1;
 	    } else {
 		# No symbol file found, fall back to standard shlibs
                 print "Using shlibs+objdump for $soname (file $lib)\n" if $debug;
@@ -269,7 +267,7 @@ foreach my $file (keys %exec) {
                 my $libobj = $objdump_cache{$lib};
                 my $id = $dumplibs_wo_symfile->add_object($libobj);
 		if (($id ne $soname) and ($id ne $lib)) {
-		    warning(_g('%s has an unexpected SONAME (%s)'), $lib, $id);
+		    warning(g_('%s has an unexpected SONAME (%s)'), $lib, $id);
 		    $alt_soname{$id} = $soname;
 		}
 		push @soname_wo_symfile, $soname;
@@ -290,7 +288,7 @@ foreach my $file (keys %exec) {
 		    $ignore++ unless scalar(split_soname($soname));
 		    # 3/ when we have been asked to do so
 		    $ignore++ if $ignore_missing_info;
-		    error(_g('no dependency information found for %s ' .
+		    error(g_('no dependency information found for %s ' .
 		             "(used by %s)\n" .
 		             'Hint: check if the library actually comes ' .
 		             'from a package.'), $lib, $file)
@@ -303,14 +301,14 @@ foreach my $file (keys %exec) {
     # Scan all undefined symbols of the binary and resolve to a
     # dependency
     my %soname_used;
-    foreach (@sonames) {
+    foreach my $soname (@sonames) {
         # Initialize statistics
-        $soname_used{$_} = 0;
-        $global_soname_used{$_} //= 0;
-        if (exists $global_soname_needed{$_}) {
-            push @{$global_soname_needed{$_}}, $file;
+        $soname_used{$soname} = 0;
+        $global_soname_used{$soname} //= 0;
+        if (exists $global_soname_needed{$soname}) {
+            push @{$global_soname_needed{$soname}}, $file;
         } else {
-            $global_soname_needed{$_} = [ $file ];
+            $global_soname_needed{$soname} = [ $file ];
         }
     }
     my $nb_warnings = 0;
@@ -323,7 +321,7 @@ foreach my $file (keys %exec) {
         my $parent_dir = '/' . dirname($relname);
         $in_public_dir = any { $parent_dir eq $_ } get_library_paths();
     } else {
-        warning(_g('binaries to analyze should already be ' .
+        warning(g_('binaries to analyze should already be ' .
                    "installed in their package's directory"));
     }
     print "Analyzing all undefined symbols\n" if $debug > 1;
@@ -366,10 +364,10 @@ foreach my $file (keys %exec) {
                             or (not $in_public_dir and $nb_warnings < 1))
                         {
                             if ($in_public_dir) {
-			        warning(_g('symbol %s used by %s found in none of the ' .
+			        warning(g_('symbol %s used by %s found in none of the ' .
 				           'libraries'), $print_name, $file);
                             } else {
-			        warning(_g('%s contains an unresolvable reference to ' .
+			        warning(g_('%s contains an unresolvable reference to ' .
                                            "symbol %s: it's probably a plugin"),
                                         $file, $print_name);
                             }
@@ -418,7 +416,7 @@ foreach my $file (keys %exec) {
 	    next if ($soname =~ /^libm\.so\.\d+$/ and
 	             any { m/^libstdc\+\+\.so\.\d+/ } @sonames);
             next unless ($warnings & WARN_NOT_NEEDED);
-	    warning(_g('%s should not be linked against %s (it uses none of ' .
+	    warning(g_('%s should not be linked against %s (it uses none of ' .
 	               "the library's symbols)"), $file, $soname);
 	}
     }
@@ -442,7 +440,7 @@ foreach my $soname (keys %global_soname_needed) {
 
 # Quit now if any missing libraries
 if ($error_count >= 1) {
-    my $note = _g('Note: libraries are not searched in other binary packages ' .
+    my $note = g_('Note: libraries are not searched in other binary packages ' .
 	"that do not have any shlibs or symbols file.\nTo help dpkg-shlibdeps " .
 	'find private libraries, you might need to use -l.');
     error(P_('cannot continue due to the error above',
@@ -451,24 +449,15 @@ if ($error_count >= 1) {
 }
 
 # Open substvars file
-my $fh;
+
+my $substvars = Dpkg::Substvars->new();
 if ($stdout) {
-    $fh = \*STDOUT;
+    $varlistfilenew = '-';
 } else {
-    open(my $new_fh, '>', "$varlistfile.new")
-        or syserr(_g("open new substvars file \`%s'"), "$varlistfile.new");
-    if (-e $varlistfile) {
-	open(my $old_fh, '<', $varlistfile)
-	    or syserr(_g("open old varlist file \`%s' for reading"), $varlistfile);
-	while (my $entry = <$old_fh>) {
-	    next if $entry =~ m/^\Q$varnameprefix\E:/;
-	    print { $new_fh } $entry
-	        or syserr(_g("copy old entry to new varlist file \`%s'"),
-	                  "$varlistfile.new");
-	}
-	close($old_fh);
-    }
-    $fh = $new_fh;
+    $substvars->load($varlistfile) if -e $varlistfile;
+    $substvars->filter(remove => sub { $_[0] =~ m/^\Q$varnameprefix\E:/ });
+
+    $varlistfilenew = "$varlistfile.new";
 }
 
 # Write out the shlibs substvars
@@ -481,22 +470,22 @@ sub filter_deps {
 	return 0 if $dep =~ /^\s*\Q$exc\E\b/;
     }
     # Don't include dependencies if they are already
-    # mentionned in a higher priority field
+    # mentioned in a higher priority field
     if (not exists($depseen{$dep})) {
 	$depseen{$dep} = $dependencies{$field}{$dep};
 	return 1;
     } else {
-	# Since dependencies can be versionned, we have to
+	# Since dependencies can be versioned, we have to
 	# verify if the dependency is stronger than the
 	# previously seen one
 	my $stronger;
 	if ($depseen{$dep} eq $dependencies{$field}{$dep}) {
-	    # If both versions are the same (possibly unversionned)
+	    # If both versions are the same (possibly unversioned)
 	    $stronger = 0;
 	} elsif ($dependencies{$field}{$dep} eq '') {
-	    $stronger = 0; # If the dep is unversionned
+	    $stronger = 0; # If the dep is unversioned
 	} elsif ($depseen{$dep} eq '') {
-	    $stronger = 1; # If the dep seen is unversionned
+	    $stronger = 1; # If the dep seen is unversioned
 	} elsif (version_compare_relation($depseen{$dep}, REL_GT,
                                           $dependencies{$field}{$dep})) {
 	    # The version of the dep seen is stronger...
@@ -515,29 +504,32 @@ foreach my $field (reverse @depfields) {
 	$dep = join ', ',
 	    map {
 		# Translate dependency templates into real dependencies
-		if ($dependencies{$field}{$_}) {
-		    s/#MINVER#/(>= $dependencies{$field}{$_})/g;
+		my $templ = $_;
+		if ($dependencies{$field}{$templ}) {
+		    $templ =~ s/#MINVER#/(>= $dependencies{$field}{$templ})/g;
 		} else {
-		    s/#MINVER#//g;
+		    $templ =~ s/#MINVER#//g;
 		}
-		s/\s+/ /g;
-		$_;
-	    } grep { filter_deps($_, $field) }
-	    keys %{$dependencies{$field}};
+		$templ =~ s/\s+/ /g;
+		$templ;
+	    } grep {
+		filter_deps($_, $field)
+	    } keys %{$dependencies{$field}};
     }
     if ($dep) {
         my $obj = deps_parse($dep);
-        error(_g('invalid dependency got generated: %s'), $dep) unless defined $obj;
+        error(g_('invalid dependency got generated: %s'), $dep) unless defined $obj;
         $obj->sort();
-	print { $fh } "$varnameprefix:$field=$obj\n";
+        $substvars->set_as_used("$varnameprefix:$field", "$obj");
     }
 }
 
+$substvars->save($varlistfilenew);
+
 # Replace old file by new one
 if (!$stdout) {
-    close($fh) or syserr(_g('cannot close %s'), "$varlistfile.new");
-    rename "$varlistfile.new", $varlistfile
-        or syserr(_g("install new varlist file \`%s'"), $varlistfile);
+    rename $varlistfilenew, $varlistfile
+        or syserr(g_("install new varlist file '%s'"), $varlistfile);
 }
 
 ##
@@ -545,23 +537,23 @@ if (!$stdout) {
 ##
 
 sub version {
-    printf _g("Debian %s version %s.\n"), $Dpkg::PROGNAME, $Dpkg::PROGVERSION;
+    printf g_("Debian %s version %s.\n"), $Dpkg::PROGNAME, $Dpkg::PROGVERSION;
 
-    printf _g('
+    printf g_('
 This is free software; see the GNU General Public License version 2 or
 later for copying conditions. There is NO warranty.
 ');
 }
 
 sub usage {
-    printf _g(
+    printf g_(
 'Usage: %s [<option>...] <executable>|-e<executable> [<option>...]')
-    . "\n\n" . _g(
+    . "\n\n" . g_(
 "Positional options (order is significant):
   <executable>             include dependencies for <executable>,
   -e<executable>           (use -e if <executable> starts with '-')
   -d<dependency-field>     next executable(s) set shlibs:<dependency-field>.")
-    . "\n\n" . _g(
+    . "\n\n" . g_(
 "Options:
   -l<library-dir>          add directory to private shared library search list.
   -p<varname-prefix>       set <varname-prefix>:* instead of shlibs:*.
@@ -578,7 +570,7 @@ sub usage {
   --admindir=<directory>   change the administrative directory.
   -?, --help               show this help message.
       --version            show the version.")
-    . "\n\n" . _g(
+    . "\n\n" . g_(
 'Dependency fields recognized are:
   %s
 '), $Dpkg::PROGNAME, join('/', @depfields);
@@ -620,7 +612,7 @@ sub update_dependency_version {
 	if (exists $dependencies{$cur_field}{$subdep} and
 	    defined($dependencies{$cur_field}{$subdep}))
 	{
-	    if ($dependencies{$cur_field}{$subdep} eq '' or
+	    if ($dependencies{$cur_field}{$subdep} eq '' or $minver ne '' and
 		version_compare_relation($minver, REL_GT,
 				         $dependencies{$cur_field}{$subdep}))
 	    {
@@ -680,22 +672,35 @@ sub split_soname {
 
 sub extract_from_shlibs {
     my ($soname, $shlibfile) = @_;
+
+    my $shlibs_re = qr{
+        ^\s*
+        (?:(\S+):\s+)?              # Optional type
+        (\S+)\s+                    # Library
+        (\S+)                       # Version
+        (?:
+          \s+
+          (\S.*\S)                  # Dependencies
+        )?
+        \s*$
+    }x;
+
     # Split soname in name/version
     my ($libname, $libversion) = split_soname($soname);
     unless (defined $libname) {
-	warning(_g("can't extract name and version from library name '%s'"),
+	warning(g_("can't extract name and version from library name '%s'"),
 	        $soname);
 	return;
     }
     # Open shlibs file
     open(my $shlibs_fh, '<', $shlibfile)
-        or syserr(_g("unable to open shared libs info file \`%s'"), $shlibfile);
+        or syserr(g_("unable to open shared libs info file '%s'"), $shlibfile);
     my $dep;
     while (<$shlibs_fh>) {
 	s/\s*\n$//;
 	next if m/^\#/;
-	if (!m/^\s*(?:(\S+):\s+)?(\S+)\s+(\S+)(?:\s+(\S.*\S))?\s*$/) {
-	    warning(_g("shared libs info file \`%s' line %d: bad line \`%s'"),
+	if (!m/$shlibs_re/) {
+	    warning(g_("shared libs info file '%s' line %d: bad line '%s'"),
 	            $shlibfile, $., $_);
 	    next;
 	}
@@ -703,7 +708,7 @@ sub extract_from_shlibs {
 	if (($libname eq $2) && ($libversion eq $3)) {
 	    # Define dep and end here if the package type explicitly
 	    # matches. Otherwise if the packagetype is not specified, use
-	    # the dep only as a default that can be overriden by a later
+	    # the dep only as a default that can be overridden by a later
 	    # line
 	    if (defined($1)) {
 		if ($1 eq $packagetype) {
@@ -754,7 +759,7 @@ sub symfile_has_soname {
     }
 
     open(my $symfile_fh, '<', $file)
-        or syserr(_g('cannot open file %s'), $file);
+        or syserr(g_('cannot open file %s'), $file);
     my $result = 0;
     while (<$symfile_fh>) {
 	if (/^\Q$soname\E /) {
@@ -788,7 +793,7 @@ sub my_find_library {
 		$path =~ s/\$ORIGIN/$origin/g;
 		$path =~ s/\$\{ORIGIN\}/$origin/g;
 	    } else {
-		warning(_g('$ORIGIN is used in RPATH of %s and the corresponding ' .
+		warning(g_('$ORIGIN is used in RPATH of %s and the corresponding ' .
 		'directory could not be identified due to lack of DEBIAN ' .
 		"sub-directory in the root of package's build tree"), $execfile);
 	    }
@@ -829,39 +834,39 @@ sub find_packages {
     my @files;
     my $pkgmatch = {};
 
-    foreach (@_) {
-	if (exists $cached_pkgmatch{$_}) {
-	    $pkgmatch->{$_} = $cached_pkgmatch{$_};
+    foreach my $path (@_) {
+	if (exists $cached_pkgmatch{$path}) {
+	    $pkgmatch->{$path} = $cached_pkgmatch{$path};
 	} else {
-	    push @files, $_;
-	    $cached_pkgmatch{$_} = ['']; # placeholder to cache misses too.
-	    $pkgmatch->{$_} = [''];        # might be replaced later on
+	    push @files, $path;
+	    $cached_pkgmatch{$path} = ['']; # placeholder to cache misses too.
+	    $pkgmatch->{$path} = [''];      # might be replaced later on
 	}
     }
     return $pkgmatch unless scalar(@files);
 
     my $pid = open(my $dpkg_fh, '-|');
-    syserr(_g('cannot fork for %s'), 'dpkg --search') unless defined($pid);
+    syserr(g_('cannot fork for %s'), 'dpkg-query --search') unless defined $pid;
     if (!$pid) {
 	# Child process running dpkg --search and discarding errors
 	close STDERR;
 	open STDERR, '>', '/dev/null'
-	    or syserr(_g('cannot open file %s'), '/dev/null');
+	    or syserr(g_('cannot open file %s'), '/dev/null');
 	$ENV{LC_ALL} = 'C';
-	exec('dpkg', '--search', '--', @files)
-	    or syserr(_g('unable to execute %s'), 'dpkg');
+	exec 'dpkg-query', '--search', '--', @files
+	    or syserr(g_('unable to execute %s'), 'dpkg');
     }
-    while (defined($_ = <$dpkg_fh>)) {
-	chomp($_);
+    while (<$dpkg_fh>) {
+	chomp;
 	if (m/^local diversion |^diversion by/) {
-	    warning(_g('diversions involved - output may be incorrect'));
+	    warning(g_('diversions involved - output may be incorrect'));
 	    print { *STDERR } " $_\n"
-		or syserr(_g('write diversion info to stderr'));
+		or syserr(g_('write diversion info to stderr'));
 	} elsif (m/^([-a-z0-9+.:, ]+): (\/.*)$/) {
 	    my ($pkgs, $path) = ($1, $2);
 	    $cached_pkgmatch{$path} = $pkgmatch->{$path} = [ split /, /, $pkgs ];
 	} else {
-	    warning(_g("unknown output from dpkg --search: '%s'"), $_);
+	    warning(g_("unknown output from dpkg --search: '%s'"), $_);
 	}
     }
     close($dpkg_fh);
