@@ -2,7 +2,8 @@
  * dpkg - main program for package management
  * packages.c - common to actions that process packages
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
+ * Copyright © 2006-2014 Guillem Jover <guillem@debian.org>
  * Copyright © 2011 Linaro Limited
  * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
@@ -27,7 +28,6 @@
 #include <sys/stat.h>
 
 #include <assert.h>
-#include <ctype.h>
 #include <string.h>
 #include <fcntl.h>
 #include <dirent.h>
@@ -40,7 +40,7 @@
 #include <dpkg/dpkg-db.h>
 #include <dpkg/pkg-list.h>
 #include <dpkg/pkg-queue.h>
-#include <dpkg/pkg-spec.h>
+#include <dpkg/string.h>
 #include <dpkg/options.h>
 
 #include "filesdb.h"
@@ -50,14 +50,26 @@
 static struct pkginfo *progress_bytrigproc;
 static struct pkg_queue queue = PKG_QUEUE_INIT;
 
-int sincenothing = 0, dependtry = 0;
+enum dependtry dependtry = DEPEND_TRY_NORMAL;
+int sincenothing = 0;
 
 void
 enqueue_package(struct pkginfo *pkg)
 {
+  ensure_package_clientdata(pkg);
+  if (pkg->clientdata->enqueued)
+    return;
+  pkg->clientdata->enqueued = true;
   pkg_queue_push(&queue, pkg);
 }
 
+void
+enqueue_package_mark_seen(struct pkginfo *pkg)
+{
+  enqueue_package(pkg);
+  pkg->clientdata->cmdline_seen++;
+}
+
 static void
 enqueue_pending(void)
 {
@@ -68,28 +80,28 @@ enqueue_pending(void)
   while ((pkg = pkg_db_iter_next_pkg(it)) != NULL) {
     switch (cipaction->arg_int) {
     case act_configure:
-      if (!(pkg->status == stat_unpacked ||
-            pkg->status == stat_halfconfigured ||
+      if (!(pkg->status == PKG_STAT_UNPACKED ||
+            pkg->status == PKG_STAT_HALFCONFIGURED ||
             pkg->trigpend_head))
         continue;
-      if (pkg->want != want_install)
+      if (pkg->want != PKG_WANT_INSTALL)
         continue;
       break;
     case act_triggers:
       if (!pkg->trigpend_head)
         continue;
-      if (pkg->want != want_install)
+      if (pkg->want != PKG_WANT_INSTALL)
         continue;
       break;
     case act_remove:
     case act_purge:
-      if (pkg->want != want_purge) {
-        if (pkg->want != want_deinstall)
+      if (pkg->want != PKG_WANT_PURGE) {
+        if (pkg->want != PKG_WANT_DEINSTALL)
           continue;
-        if (pkg->status == stat_configfiles)
+        if (pkg->status == PKG_STAT_CONFIGFILES)
           continue;
       }
-      if (pkg->status == stat_notinstalled)
+      if (pkg->status == PKG_STAT_NOTINSTALLED)
         continue;
       break;
     default:
@@ -106,24 +118,19 @@ enqueue_specified(const char *const *argv)
   const char *thisarg;
 
   while ((thisarg = *argv++) != NULL) {
-    struct dpkg_error err;
     struct pkginfo *pkg;
 
-    pkg = pkg_spec_parse_pkg(thisarg, &err);
-    if (pkg == NULL)
-      badusage(_("--%s needs a valid package name but '%.250s' is not: %s"),
-               cipaction->olong, thisarg, err.str);
-
-    if (pkg->status == stat_notinstalled) {
-      size_t l = strlen(pkg->set->name);
-      const char *extension = pkg->set->name + l - sizeof(DEBEXT) + 1;
-
-      if (l >= sizeof(DEBEXT) && strcmp(extension, DEBEXT) == 0)
-        badusage(_("you must specify packages by their own names,"
-                   " not by quoting the names of the files they come in"));
+    pkg = dpkg_options_parse_pkgname(cipaction, thisarg);
+    if (pkg->status == PKG_STAT_NOTINSTALLED &&
+        str_match_end(pkg->set->name, DEBEXT)) {
+      badusage(_("you must specify packages by their own names, "
+                 "not by quoting the names of the files they come in"));
     }
-    enqueue_package(pkg);
+    enqueue_package_mark_seen(pkg);
   }
+
+  if (cipaction->arg_int == act_configure)
+    trigproc_populate_deferred();
 }
 
 int
@@ -166,7 +173,7 @@ void process_queue(void) {
   struct pkginfo *volatile pkg;
   volatile enum action action_todo;
   jmp_buf ejbuf;
-  enum istobes istobe= itb_normal;
+  enum pkg_istobe istobe = PKG_ISTOBE_NORMAL;
 
   if (abort_processing)
     return;
@@ -175,15 +182,23 @@ void process_queue(void) {
 
   switch (cipaction->arg_int) {
   case act_triggers:
-  case act_configure: case act_install:  istobe= itb_installnew;  break;
-  case act_remove: case act_purge:       istobe= itb_remove;      break;
+  case act_configure:
+  case act_install:
+    istobe = PKG_ISTOBE_INSTALLNEW;
+    break;
+  case act_remove:
+  case act_purge:
+    istobe = PKG_ISTOBE_REMOVE;
+    break;
   default:
     internerr("unknown action '%d'", cipaction->arg_int);
   }
   for (rundown = queue.head; rundown; rundown = rundown->next) {
     ensure_package_clientdata(rundown->pkg);
-    if (rundown->pkg->clientdata->istobe == istobe) {
-      /* Erase the queue entry - this is a second copy! */
+
+    /* We have processed this package more than once. There are no duplicates
+     * as we make sure of that when enqueuing them. */
+    if (rundown->pkg->clientdata->cmdline_seen > 1) {
       switch (cipaction->arg_int) {
       case act_triggers:
       case act_configure: case act_remove: case act_purge:
@@ -198,10 +213,8 @@ void process_queue(void) {
       default:
         internerr("unknown action '%d'", cipaction->arg_int);
       }
-      rundown->pkg = NULL;
-   } else {
-      rundown->pkg->clientdata->istobe= istobe;
     }
+    rundown->pkg->clientdata->istobe = istobe;
   }
 
   while (!pkg_queue_is_empty(&queue)) {
@@ -209,26 +222,46 @@ void process_queue(void) {
     if (!pkg)
       continue; /* Duplicate, which we removed earlier. */
 
+    ensure_package_clientdata(pkg);
+    pkg->clientdata->enqueued = false;
+
     action_todo = cipaction->arg_int;
 
-    if (sincenothing++ > queue.length * 2 + 2) {
-      if (progress_bytrigproc && progress_bytrigproc->trigpend_head) {
+    if (sincenothing++ > queue.length * 3 + 2) {
+      /* Make sure that even if we have exceeded the queue since not having
+       * made any progress, we are not getting stuck trying to progress by
+       * trigger processing, w/o jumping into the next dependtry. */
+      dependtry++;
+      sincenothing = 0;
+      if (dependtry >= DEPEND_TRY_LAST)
+        internerr("exceeded dependtry %d (sincenothing=%d; queue.length=%d)",
+                  dependtry, sincenothing, queue.length);
+    } else if (sincenothing > queue.length * 2 + 2) {
+      if (dependtry >= DEPEND_TRY_TRIGGERS &&
+          progress_bytrigproc && progress_bytrigproc->trigpend_head) {
         enqueue_package(pkg);
         pkg = progress_bytrigproc;
+        progress_bytrigproc = NULL;
         action_todo = act_configure;
       } else {
         dependtry++;
         sincenothing = 0;
-        assert(dependtry <= 4);
+        if (dependtry >= DEPEND_TRY_LAST)
+          internerr("exceeded dependtry %d (sincenothing=%d, queue.length=%d)",
+                    dependtry, sincenothing, queue.length);
       }
     }
 
-    assert(pkg->status <= stat_installed);
+    debug(dbg_general, "process queue pkg %s queue.len %d progress %d, try %d",
+          pkg_name(pkg, pnaw_always), queue.length, sincenothing, dependtry);
+
+    if (pkg->status > PKG_STAT_INSTALLED)
+      internerr("package status (%d) > PKG_STAT_INSTALLED", pkg->status);
 
     if (setjmp(ejbuf)) {
       /* Give up on it from the point of view of other packages, i.e. reset
        * istobe. */
-      pkg->clientdata->istobe= itb_normal;
+      pkg->clientdata->istobe = PKG_ISTOBE_NORMAL;
 
       pop_error_context(ehflag_bombout);
       if (abort_processing)
@@ -242,18 +275,18 @@ void process_queue(void) {
     case act_triggers:
       if (!pkg->trigpend_head)
         ohshit(_("package %.250s is not ready for trigger processing\n"
-                 " (current status `%.250s' with no pending triggers)"),
-               pkg_name(pkg, pnaw_nonambig), statusinfos[pkg->status].name);
+                 " (current status '%.250s' with no pending triggers)"),
+               pkg_name(pkg, pnaw_nonambig), pkg_status_name(pkg));
       /* Fall through. */
     case act_install:
       /* Don't try to configure pkgs that we've just disappeared. */
-      if (pkg->status == stat_notinstalled)
+      if (pkg->status == PKG_STAT_NOTINSTALLED)
         break;
       /* Fall through. */
     case act_configure:
       /* Do whatever is most needed. */
       if (pkg->trigpend_head)
-        trigproc(pkg);
+        trigproc(pkg, TRIGPROC_TRY_QUEUED);
       else
         deferred_configure(pkg);
       break;
@@ -309,12 +342,21 @@ void process_queue(void) {
  */
 
 enum found_status {
-  found_none = 0,
-  found_defer = 1,
-  found_forced = 2,
-  found_ok = 3,
+  FOUND_NONE = 0,
+  FOUND_DEFER = 1,
+  FOUND_FORCED = 2,
+  FOUND_OK = 3,
 };
 
+static enum found_status
+found_forced_on(enum dependtry dependtry_forced)
+{
+  if (dependtry >= dependtry_forced)
+    return FOUND_FORCED;
+  else
+    return FOUND_DEFER;
+}
+
 /*
  * Return values:
  *   0: cannot be satisfied.
@@ -328,7 +370,7 @@ enum found_status {
  */
 static enum found_status
 deppossi_ok_found(struct pkginfo *possdependee, struct pkginfo *requiredby,
-                  struct pkginfo *removing, struct pkgset *providing,
+                  struct pkginfo *removing, struct deppossi *provider,
                   struct pkginfo **fixbytrig,
                   bool *matched, struct deppossi *checkversion,
                   int *interestingwarnings, struct varbuf *oemsgs)
@@ -337,14 +379,15 @@ deppossi_ok_found(struct pkginfo *possdependee, struct pkginfo *requiredby,
 
   if (ignore_depends(possdependee)) {
     debug(dbg_depcondetail,"      ignoring depended package so ok and found");
-    return found_ok;
+    return FOUND_OK;
   }
-  thisf = found_none;
+  thisf = FOUND_NONE;
   if (possdependee == removing) {
-    if (providing) {
+    if (provider) {
       varbuf_printf(oemsgs,
                     _("  Package %s which provides %s is to be removed.\n"),
-                    pkg_name(possdependee, pnaw_nonambig), providing->name);
+                    pkg_name(possdependee, pnaw_nonambig),
+                    provider->ed->name);
     } else {
       varbuf_printf(oemsgs, _("  Package %s is to be removed.\n"),
                     pkg_name(possdependee, pnaw_nonambig));
@@ -355,36 +398,56 @@ deppossi_ok_found(struct pkginfo *possdependee, struct pkginfo *requiredby,
     return thisf;
   }
   switch (possdependee->status) {
-  case stat_unpacked:
-  case stat_halfconfigured:
-  case stat_triggersawaited:
-  case stat_triggerspending:
-  case stat_installed:
-    if (checkversion && !versionsatisfied(&possdependee->installed,checkversion)) {
-      varbuf_printf(oemsgs, _("  Version of %s on system is %s.\n"),
-                    pkg_name(possdependee, pnaw_nonambig),
-                    versiondescribe(&possdependee->installed.version,
-                                    vdew_nonambig));
-      assert(checkversion->verrel != dpkg_relation_none);
-      if (fc_dependsversion)
-        thisf = (dependtry >= 3) ? found_forced : found_defer;
-      debug(dbg_depcondetail,"      bad version, returning %d",thisf);
-      (*interestingwarnings)++;
-      return thisf;
+  case PKG_STAT_UNPACKED:
+  case PKG_STAT_HALFCONFIGURED:
+  case PKG_STAT_TRIGGERSAWAITED:
+  case PKG_STAT_TRIGGERSPENDING:
+  case PKG_STAT_INSTALLED:
+    if (checkversion) {
+      if (provider) {
+        debug(dbg_depcondetail, "      checking package %s provided by pkg %s",
+              checkversion->ed->name, pkg_name(possdependee, pnaw_always));
+        if (!pkg_virtual_deppossi_satisfied(checkversion, provider)) {
+          varbuf_printf(oemsgs,
+                        _("  Version of %s on system, provided by %s, is %s.\n"),
+                        checkversion->ed->name,
+                        pkg_name(possdependee, pnaw_always),
+                        versiondescribe(&provider->version, vdew_nonambig));
+          if (fc_dependsversion)
+            thisf = found_forced_on(DEPEND_TRY_FORCE_DEPENDS_VERSION);
+          debug(dbg_depcondetail, "      bad version");
+          goto unsuitable;
+        }
+      } else {
+        debug(dbg_depcondetail, "      checking non-provided pkg %s",
+              pkg_name(possdependee, pnaw_always));
+        if (!versionsatisfied(&possdependee->installed, checkversion)) {
+          varbuf_printf(oemsgs, _("  Version of %s on system is %s.\n"),
+                        pkg_name(possdependee, pnaw_nonambig),
+                        versiondescribe(&possdependee->installed.version,
+                                        vdew_nonambig));
+          if (fc_dependsversion)
+            thisf = found_forced_on(DEPEND_TRY_FORCE_DEPENDS_VERSION);
+          debug(dbg_depcondetail, "      bad version");
+          goto unsuitable;
+        }
+      }
     }
-    if (possdependee->status == stat_installed ||
-        possdependee->status == stat_triggerspending) {
+    if (possdependee->status == PKG_STAT_INSTALLED ||
+        possdependee->status == PKG_STAT_TRIGGERSPENDING) {
       debug(dbg_depcondetail,"      is installed, ok and found");
-      return found_ok;
+      return FOUND_OK;
     }
-    if (possdependee->status == stat_triggersawaited) {
+    if (possdependee->status == PKG_STAT_TRIGGERSAWAITED) {
       assert(possdependee->trigaw.head);
-      if (removing || !(f_triggers ||
-                        possdependee->clientdata->istobe == itb_installnew)) {
-        if (providing) {
+      if (removing ||
+          !(f_triggers ||
+            possdependee->clientdata->istobe == PKG_ISTOBE_INSTALLNEW)) {
+        if (provider) {
           varbuf_printf(oemsgs,
                         _("  Package %s which provides %s awaits trigger processing.\n"),
-                        pkg_name(possdependee, pnaw_nonambig), providing->name);
+                        pkg_name(possdependee, pnaw_nonambig),
+                        provider->ed->name);
         } else {
           varbuf_printf(oemsgs,
                         _("  Package %s awaits trigger processing.\n"),
@@ -407,26 +470,27 @@ deppossi_ok_found(struct pkginfo *possdependee, struct pkginfo *requiredby,
       debug(dbg_depcondetail,
             "      triggers-awaited, fixbytrig '%s', defer",
             pkg_name(*fixbytrig, pnaw_always));
-      return found_defer;
+      return FOUND_DEFER;
     }
     if (possdependee->clientdata &&
-        possdependee->clientdata->istobe == itb_installnew) {
+        possdependee->clientdata->istobe == PKG_ISTOBE_INSTALLNEW) {
       debug(dbg_depcondetail,"      unpacked/halfconfigured, defer");
-      return found_defer;
+      return FOUND_DEFER;
     } else if (!removing && fc_configureany &&
                !skip_due_to_hold(possdependee) &&
-               !(possdependee->status == stat_halfconfigured)) {
+               !(possdependee->status == PKG_STAT_HALFCONFIGURED)) {
       notice(_("also configuring '%s' (required by '%s')"),
              pkg_name(possdependee, pnaw_nonambig),
              pkg_name(requiredby, pnaw_nonambig));
       enqueue_package(possdependee);
       sincenothing = 0;
-      return found_defer;
+      return FOUND_DEFER;
     } else {
-      if (providing) {
+      if (provider) {
         varbuf_printf(oemsgs,
                       _("  Package %s which provides %s is not configured yet.\n"),
-                      pkg_name(possdependee, pnaw_nonambig), providing->name);
+                      pkg_name(possdependee, pnaw_nonambig),
+                      provider->ed->name);
       } else {
         varbuf_printf(oemsgs, _("  Package %s is not configured yet.\n"),
                       pkg_name(possdependee, pnaw_nonambig));
@@ -437,10 +501,11 @@ deppossi_ok_found(struct pkginfo *possdependee, struct pkginfo *requiredby,
     }
 
   default:
-    if (providing) {
+    if (provider) {
       varbuf_printf(oemsgs,
                     _("  Package %s which provides %s is not installed.\n"),
-                    pkg_name(possdependee, pnaw_nonambig), providing->name);
+                    pkg_name(possdependee, pnaw_nonambig),
+                    provider->ed->name);
     } else {
       varbuf_printf(oemsgs, _("  Package %s is not installed.\n"),
                     pkg_name(possdependee, pnaw_nonambig));
@@ -460,16 +525,17 @@ unsuitable:
 static void
 breaks_check_one(struct varbuf *aemsgs, enum dep_check *ok,
                  struct deppossi *breaks, struct pkginfo *broken,
-                 struct pkginfo *breaker, struct pkgset *virtbroken)
+                 struct pkginfo *breaker, struct deppossi *virtbroken)
 {
   struct varbuf depmsg = VARBUF_INIT;
 
   debug(dbg_depcondetail, "      checking breaker %s virtbroken %s",
         pkg_name(breaker, pnaw_always),
-        virtbroken ? virtbroken->name : "<none>");
+        virtbroken ? virtbroken->ed->name : "<none>");
 
-  if (breaker->status == stat_notinstalled ||
-      breaker->status == stat_configfiles) return;
+  if (breaker->status == PKG_STAT_NOTINSTALLED ||
+      breaker->status == PKG_STAT_CONFIGFILES)
+    return;
   if (broken == breaker) return;
   if (!versionsatisfied(&broken->installed, breaks)) return;
   /* The test below can only trigger if dep_breaks start having
@@ -477,7 +543,9 @@ breaks_check_one(struct varbuf *aemsgs, enum dep_check *ok,
   if (!archsatisfied(&broken->installed, breaks))
     return;
   if (ignore_depends(breaker)) return;
-  if (virtbroken && ignore_depends(&virtbroken->pkg))
+  if (virtbroken && ignore_depends(&virtbroken->ed->pkg))
+    return;
+  if (virtbroken && !pkg_virtual_deppossi_satisfied(breaks, virtbroken))
     return;
 
   varbufdependency(&depmsg, breaks->up);
@@ -492,28 +560,26 @@ breaks_check_one(struct varbuf *aemsgs, enum dep_check *ok,
     varbuf_printf(aemsgs, _("  %s (%s) provides %s.\n"),
                   pkg_name(broken, pnaw_nonambig),
                   versiondescribe(&broken->installed.version, vdew_nonambig),
-                  virtbroken->name);
-  } else if (breaks->verrel != dpkg_relation_none) {
+                  virtbroken->ed->name);
+  } else if (breaks->verrel != DPKG_RELATION_NONE) {
     varbuf_printf(aemsgs, _("  Version of %s to be configured is %s.\n"),
                   pkg_name(broken, pnaw_nonambig),
                   versiondescribe(&broken->installed.version, vdew_nonambig));
     if (fc_dependsversion) return;
   }
   if (force_breaks(breaks)) return;
-  *ok = dep_check_halt;
+  *ok = DEP_CHECK_HALT;
 }
 
 static void
 breaks_check_target(struct varbuf *aemsgs, enum dep_check *ok,
                     struct pkginfo *broken, struct pkgset *target,
-                    struct pkgset *virtbroken)
+                    struct deppossi *virtbroken)
 {
   struct deppossi *possi;
 
   for (possi = target->depended.installed; possi; possi = possi->rev_next) {
     if (possi->up->type != dep_breaks) continue;
-    if (virtbroken && possi->verrel != dpkg_relation_none)
-      continue;
     breaks_check_one(aemsgs, ok, possi, broken, possi->up->up, virtbroken);
   }
 }
@@ -522,8 +588,8 @@ enum dep_check
 breakses_ok(struct pkginfo *pkg, struct varbuf *aemsgs)
 {
   struct dependency *dep;
-  struct pkgset *virtbroken;
-  enum dep_check ok = dep_check_ok;
+  struct deppossi *virtbroken;
+  enum dep_check ok = DEP_CHECK_OK;
 
   debug(dbg_depcon, "    checking Breaks");
 
@@ -531,9 +597,9 @@ breakses_ok(struct pkginfo *pkg, struct varbuf *aemsgs)
 
   for (dep= pkg->installed.depends; dep; dep= dep->next) {
     if (dep->type != dep_provides) continue;
-    virtbroken = dep->list->ed;
-    debug(dbg_depcondetail, "     checking virtbroken %s", virtbroken->name);
-    breaks_check_target(aemsgs, &ok, pkg, virtbroken, virtbroken);
+    virtbroken = dep->list;
+    debug(dbg_depcondetail, "     checking virtbroken %s", virtbroken->ed->name);
+    breaks_check_target(aemsgs, &ok, pkg, virtbroken->ed, virtbroken);
   }
   return ok;
 }
@@ -557,7 +623,7 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
   struct pkginfo *possfixbytrig, *canfixbytrig;
 
   interestingwarnings= 0;
-  ok = dep_check_ok;
+  ok = DEP_CHECK_OK;
   debug(dbg_depcon,"checking dependencies of %s (- %s)",
         pkg_name(pkg, pnaw_always),
         removing ? pkg_name(removing, pnaw_always) : "<none>");
@@ -569,20 +635,20 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
     debug(dbg_depcondetail,"  checking group ...");
     matched = false;
     varbuf_reset(&oemsgs);
-    found = found_none;
+    found = FOUND_NONE;
     possfixbytrig = NULL;
-    for (possi = dep->list; found != found_ok && possi; possi = possi->next) {
+    for (possi = dep->list; found != FOUND_OK && possi; possi = possi->next) {
       struct deppossi_pkg_iterator *possi_iter;
       struct pkginfo *pkg_pos;
 
       debug(dbg_depcondetail,"    checking possibility  -> %s",possi->ed->name);
       if (possi->cyclebreak) {
         debug(dbg_depcondetail,"      break cycle so ok and found");
-        found = found_ok;
+        found = FOUND_OK;
         break;
       }
 
-      thisf = found_none;
+      thisf = FOUND_NONE;
       possi_iter = deppossi_pkg_iter_new(possi, wpb_installed);
       while ((pkg_pos = deppossi_pkg_iter_next(possi_iter))) {
         thisf = deppossi_ok_found(pkg_pos, pkg, removing, NULL,
@@ -590,14 +656,14 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
                                   &interestingwarnings, &oemsgs);
         if (thisf > found)
           found = thisf;
-        if (found == found_ok)
+        if (found == FOUND_OK)
           break;
       }
       deppossi_pkg_iter_free(possi_iter);
 
-      if (found != found_ok && possi->verrel == dpkg_relation_none) {
+      if (found != FOUND_OK) {
         for (provider = possi->ed->depended.installed;
-             found != found_ok && provider;
+             found != FOUND_OK && provider;
              provider = provider->rev_next) {
           if (provider->up->type != dep_provides)
             continue;
@@ -608,10 +674,15 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
             debug(dbg_depcondetail, "       provider does not satisfy arch");
             continue;
           }
-          thisf = deppossi_ok_found(provider->up->up, pkg, removing,
-                                    possi->ed,
-                                    &possfixbytrig, &matched, NULL,
+          thisf = deppossi_ok_found(provider->up->up, pkg, removing, provider,
+                                    &possfixbytrig, &matched, possi,
                                     &interestingwarnings, &oemsgs);
+          if (thisf == FOUND_DEFER && provider->up->up == pkg && !removing) {
+            /* IOW, if the pkg satisfies its own dep (via a provide), then
+             * we let it pass, even if it isn't configured yet (as we're
+             * installing it). */
+             thisf = FOUND_OK;
+          }
           if (thisf > found)
             found = thisf;
         }
@@ -620,7 +691,7 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
       if (thisf > found) found= thisf;
     }
     if (fc_depends) {
-      thisf = (dependtry >= 4) ? found_forced : found_defer;
+      thisf = found_forced_on(DEPEND_TRY_FORCE_DEPENDS);
       if (thisf > found) {
         found = thisf;
         debug(dbg_depcondetail, "  rescued by force-depends, found %d", found);
@@ -631,10 +702,11 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
           possfixbytrig ? pkg_name(possfixbytrig, pnaw_always) : "-");
     if (removing && !matched) continue;
     switch (found) {
-    case found_none:
+    case FOUND_NONE:
       anycannotfixbytrig = true;
-      ok = dep_check_halt;
-    case found_forced:
+      ok = DEP_CHECK_HALT;
+      /* Fall through. */
+    case FOUND_FORCED:
       varbuf_add_str(aemsgs, " ");
       varbuf_add_pkgbin_name(aemsgs, pkg, &pkg->installed, pnaw_nonambig);
       varbuf_add_str(aemsgs, _(" depends on "));
@@ -649,23 +721,23 @@ dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
         varbuf_add_str(aemsgs, ".\n");
       }
       break;
-    case found_defer:
+    case FOUND_DEFER:
       if (possfixbytrig)
         canfixbytrig = possfixbytrig;
       else
         anycannotfixbytrig = true;
-      if (ok > dep_check_defer)
-        ok = dep_check_defer;
+      if (ok > DEP_CHECK_DEFER)
+        ok = DEP_CHECK_DEFER;
       break;
-    case found_ok:
+    case FOUND_OK:
       break;
     default:
       internerr("unknown value for found '%d'", found);
     }
   }
-  if (ok == dep_check_halt &&
-      (pkg->clientdata && pkg->clientdata->istobe == itb_remove))
-    ok = dep_check_defer;
+  if (ok == DEP_CHECK_HALT &&
+      (pkg->clientdata && pkg->clientdata->istobe == PKG_ISTOBE_REMOVE))
+    ok = DEP_CHECK_DEFER;
   if (!anycannotfixbytrig && canfixbytrig)
     progress_bytrigproc = canfixbytrig;
 
