@@ -25,7 +25,7 @@
 
 #include <dpkg/macros.h>
 
-#if defined(linux) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__))
+#if defined(linux)
 #  define OSLinux
 #elif defined(__GNU__)
 #  define OSHurd
@@ -35,50 +35,45 @@
 #  define OSOpenBSD
 #elif defined(hpux)
 #  define OShpux
-#elif defined(__FreeBSD__)
+#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
 #  define OSFreeBSD
 #elif defined(__NetBSD__)
 #  define OSNetBSD
+#elif defined(__DragonFly__)
+#  define OSDragonFlyBSD
 #else
 #  error Unknown architecture - cannot build start-stop-daemon
 #endif
 
+#ifdef HAVE_SYS_PARAM_H
+#include <sys/param.h>
+#endif
 #ifdef HAVE_SYS_SYSCALL_H
 #include <sys/syscall.h>
 #endif
-
-#if defined(OSHurd)
-#include <hurd.h>
-#include <ps.h>
-#endif
-
-#if defined(OSOpenBSD) || defined(OSFreeBSD) || defined(OSNetBSD)
-#include <sys/param.h>
-#include <sys/proc.h>
-
-#include <err.h>
-#endif
-
-#ifdef HAVE_KVM_H
+#ifdef HAVE_SYS_SYSCTL_H
 #include <sys/sysctl.h>
-#include <sys/user.h>
-
-#include <kvm.h>
 #endif
-
-#if defined(OShpux)
-#include <sys/param.h>
+#ifdef HAVE_SYS_PROC_H
+#include <sys/proc.h>
+#endif
+#ifdef HAVE_SYS_USER_H
+#include <sys/user.h>
+#endif
+#ifdef HAVE_SYS_PSTAT_H
 #include <sys/pstat.h>
 #endif
-
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
+#include <sys/select.h>
 #include <sys/ioctl.h>
 
 #include <assert.h>
 #include <errno.h>
 #include <limits.h>
+#include <time.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <ctype.h>
@@ -96,10 +91,26 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <getopt.h>
-
 #ifdef HAVE_ERROR_H
 #include <error.h>
 #endif
+#ifdef HAVE_ERR_H
+#include <err.h>
+#endif
+
+#if defined(OSHurd)
+#include <hurd.h>
+#include <ps.h>
+#endif
+
+#ifdef HAVE_KVM_H
+#include <kvm.h>
+#if defined(OSFreeBSD)
+#define KVM_MEMFILE "/dev/null"
+#else
+#define KVM_MEMFILE NULL
+#endif
+#endif
 
 #ifdef _POSIX_PRIORITY_SCHEDULING
 #include <sched.h>
@@ -122,10 +133,11 @@
 #define PROCESS_NAME_SIZE 16
 #elif defined(OSFreeBSD)
 #define PROCESS_NAME_SIZE 19
+#elif defined(OSDragonFlyBSD)
+/* On DragonFlyBSD MAXCOMLEN expands to 16. */
+#define PROCESS_NAME_SIZE MAXCOMLEN
 #endif
 
-#define MIN_POLL_INTERVAL 20000 /* µs */
-
 #if defined(SYS_ioprio_set) && defined(linux)
 #define HAVE_IOPRIO_SET
 #endif
@@ -149,12 +161,22 @@ enum {
 };
 
 enum action_code {
-	action_none,
-	action_start,
-	action_stop,
-	action_status,
+	ACTION_NONE,
+	ACTION_START,
+	ACTION_STOP,
+	ACTION_STATUS,
 };
 
+/* Time conversion constants. */
+enum {
+	NANOSEC_IN_SEC      = 1000000000L,
+	NANOSEC_IN_MILLISEC = 1000000L,
+	NANOSEC_IN_MICROSEC = 1000L,
+};
+
+/* The minimum polling interval, 20ms. */
+static const long MIN_POLL_INTERVAL = 20 * NANOSEC_IN_MILLISEC;
+
 static enum action_code action;
 static bool testmode = false;
 static int quietmode = 0;
@@ -162,6 +184,7 @@ static int exitnodo = 1;
 static bool background = false;
 static bool close_io = true;
 static bool mpidfile = false;
+static bool rpidfile = false;
 static int signal_nr = SIGTERM;
 static int user_id = -1;
 static int runas_uid = -1;
@@ -174,6 +197,8 @@ static const char *changedir = "/";
 static const char *cmdname = NULL;
 static char *execname = NULL;
 static char *startas = NULL;
+static pid_t match_pid = -1;
+static pid_t match_ppid = -1;
 static const char *pidfile = NULL;
 static char what_stop[1024];
 static const char *progname = "";
@@ -187,11 +212,11 @@ static struct proc_stat_list *procset = NULL;
 
 /* LSB Init Script process status exit codes. */
 enum status_code {
-	status_ok = 0,
-	status_dead_pidfile = 1,
-	status_dead_lockfile = 2,
-	status_dead = 3,
-	status_unknown = 4,
+	STATUS_OK = 0,
+	STATUS_DEAD_PIDFILE = 1,
+	STATUS_DEAD_LOCKFILE = 2,
+	STATUS_DEAD = 3,
+	STATUS_UNKNOWN = 4,
 };
 
 struct pid_list {
@@ -254,8 +279,8 @@ fatal(const char *format, ...)
 	else
 		fprintf(stderr, "\n");
 
-	if (action == action_status)
-		exit(status_unknown);
+	if (action == ACTION_STATUS)
+		exit(STATUS_UNKNOWN);
 	else
 		exit(2);
 }
@@ -272,30 +297,58 @@ xmalloc(int size)
 }
 
 static char *
-xstrdup(const char *str)
+xstrndup(const char *str, size_t n)
 {
 	char *new_str;
 
-	new_str = strdup(str);
+	new_str = strndup(str, n);
 	if (new_str)
 		return new_str;
-	fatal("strdup(%s) failed", str);
+	fatal("strndup(%s, %zu) failed", str, n);
 }
 
 static void
-xgettimeofday(struct timeval *tv)
+timespec_gettime(struct timespec *ts)
 {
-	if (gettimeofday(tv, NULL) != 0)
+#if defined(_POSIX_TIMERS) && _POSIX_TIMERS > 0 && \
+    defined(_POSIX_MONOTONIC_CLOCK) && _POSIX_MONOTONIC_CLOCK > 0
+	if (clock_gettime(CLOCK_MONOTONIC, ts) < 0)
+		fatal("clock_gettime failed");
+#else
+	struct timeval tv;
+
+	if (gettimeofday(&tv, NULL) != 0)
 		fatal("gettimeofday failed");
+
+	ts->tv_sec = tv.tv_sec;
+	ts->tv_nsec = tv.tv_usec * NANOSEC_IN_MICROSEC;
+#endif
+}
+
+#define timespec_cmp(a, b, OP) \
+	(((a)->tv_sec == (b)->tv_sec) ? \
+	 ((a)->tv_nsec OP (b)->tv_nsec) : \
+	 ((a)->tv_sec OP (b)->tv_sec))
+
+static void
+timespec_sub(struct timespec *a, struct timespec *b, struct timespec *res)
+{
+	res->tv_sec = a->tv_sec - b->tv_sec;
+	res->tv_nsec = a->tv_nsec - b->tv_nsec;
+	if (res->tv_nsec < 0) {
+		res->tv_sec--;
+		res->tv_nsec += NANOSEC_IN_SEC;
+	}
 }
 
 static void
-tmul(struct timeval *a, int b)
+timespec_mul(struct timespec *a, int b)
 {
+	long nsec = a->tv_nsec * b;
+
 	a->tv_sec *= b;
-	a->tv_usec *= b;
-	a->tv_sec = a->tv_sec + a->tv_usec / 1000000;
-	a->tv_usec %= 1000000;
+	a->tv_sec += nsec / NANOSEC_IN_SEC;
+	a->tv_nsec = nsec % NANOSEC_IN_SEC;
 }
 
 static char *
@@ -344,7 +397,9 @@ detach_controlling_tty(void)
 static pid_t
 setsid(void)
 {
-	setpgid(0, 0);
+	if (setpgid(0, 0) < 0)
+		return -1:
+
 	detach_controlling_tty();
 
 	return 0;
@@ -352,30 +407,30 @@ setsid(void)
 #endif
 
 static void
-daemonize(void)
+wait_for_child(pid_t pid)
 {
-	pid_t pid;
+	pid_t child;
+	int status;
 
-	if (quietmode < 0)
-		printf("Detaching to start %s...", startas);
+	do {
+		child = waitpid(pid, &status, 0);
+	} while (child == -1 && errno == EINTR);
 
-	pid = fork();
-	if (pid < 0)
-		fatal("unable to do first fork");
-	else if (pid) /* Parent. */
-		_exit(0);
+	if (child != pid)
+		fatal("error waiting for child");
 
-	/* Create a new session. */
-	setsid();
+	if (WIFEXITED(status)) {
+		int err = WEXITSTATUS(status);
 
-	pid = fork();
-	if (pid < 0)
-		fatal("unable to do second fork");
-	else if (pid) /* Parent. */
-		_exit(0);
+		if (err != 0)
+			fatal("child returned error exit status %d", err);
+	} else if (WIFSIGNALED(status)) {
+		int signo = WTERMSIG(status);
 
-	if (quietmode < 0)
-		printf("done.\n");
+		fatal("child was killed by signal %d", signo);
+	} else {
+		fatal("unexpected status %d waiting for child", status);
+	}
 }
 
 static void
@@ -399,6 +454,69 @@ write_pidfile(const char *filename, pid_t pid)
 		fatal("unable to close pidfile '%s'", filename);
 }
 
+static void
+remove_pidfile(const char *filename)
+{
+	if (unlink(filename) < 0 && errno != ENOENT)
+		fatal("cannot remove pidfile '%s'", filename);
+}
+
+static void
+daemonize(void)
+{
+	pid_t pid;
+	sigset_t mask;
+	sigset_t oldmask;
+
+	if (quietmode < 0)
+		printf("Detaching to start %s...", startas);
+
+	/* Block SIGCHLD to allow waiting for the child process while it is
+	 * performing actions, such as creating a pidfile. */
+	sigemptyset(&mask);
+	sigaddset(&mask, SIGCHLD);
+	if (sigprocmask(SIG_BLOCK, &mask, &oldmask) == -1)
+		fatal("cannot block SIGCHLD");
+
+	pid = fork();
+	if (pid < 0)
+		fatal("unable to do first fork");
+	else if (pid) { /* First Parent. */
+		/* Wait for the second parent to exit, so that if we need to
+		 * perform any actions there, like creating a pidfile, we do
+		 * not suffer from race conditions on return. */
+		wait_for_child(pid);
+
+		_exit(0);
+	}
+
+	/* Create a new session. */
+	if (setsid() < 0)
+		fatal("cannot set session ID");
+
+	pid = fork();
+	if (pid < 0)
+		fatal("unable to do second fork");
+	else if (pid) { /* Second parent. */
+		/* Set a default umask for dumb programs, which might get
+		 * overridden by the --umask option later on, so that we get
+		 * a defined umask when creating the pidfille. */
+		umask(022);
+
+		if (mpidfile && pidfile != NULL)
+			/* User wants _us_ to make the pidfile. */
+			write_pidfile(pidfile, pid);
+
+		_exit(0);
+	}
+
+	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) == -1)
+		fatal("cannot restore signal mask");
+
+	if (quietmode < 0)
+		printf("done.\n");
+}
+
 static void
 pid_list_push(struct pid_list **list, pid_t pid)
 {
@@ -437,6 +555,8 @@ usage(void)
 "  -V|--version                  print version\n"
 "\n"
 "Matching options (at least one is required):\n"
+"     --pid <pid>                pid to check\n"
+"     --ppid <ppid>              parent pid to check\n"
 "  -p|--pidfile <pid-file>       pid file to check\n"
 "  -x|--exec <executable>        program to start/check if it is running\n"
 "  -n|--name <process-name>      process name to check\n"
@@ -461,6 +581,7 @@ usage(void)
 "  -b|--background               force the process to detach\n"
 "  -C|--no-close                 do not close any file descriptor\n"
 "  -m|--make-pidfile             create the pidfile before starting\n"
+"    |--remove-pidfile           delete the pidfile after stopping\n"
 "  -R|--retry <schedule>         check whether processes die, and retry\n"
 "  -t|--test                     test mode, don't do anything\n"
 "  -o|--oknodo                   exit status 0 (not 1) if nothing done\n"
@@ -506,8 +627,8 @@ badusage(const char *msg)
 		fprintf(stderr, "%s: %s\n", progname, msg);
 	fprintf(stderr, "Try '%s --help' for more information.\n", progname);
 
-	if (action == action_status)
-		exit(status_unknown);
+	if (action == ACTION_STATUS)
+		exit(STATUS_UNKNOWN);
 	else
 		exit(3);
 }
@@ -559,6 +680,17 @@ parse_unsigned(const char *string, int base, int *value_r)
 	return 0;
 }
 
+static int
+parse_pid(const char *pid_str, int *pid_num)
+{
+	if (parse_unsigned(pid_str, 10, pid_num) != 0)
+		return -1;
+	if (*pid_num == 0)
+		return -1;
+
+	return 0;
+}
+
 static int
 parse_signal(const char *sig_str, int *sig_num)
 {
@@ -601,14 +733,15 @@ validate_proc_schedule(void)
 static void
 parse_proc_schedule(const char *string)
 {
-	char *policy_str, *prio_str;
+	char *policy_str;
+	size_t policy_len;
 	int prio = 0;
 
-	policy_str = xstrdup(string);
-	policy_str = strtok(policy_str, ":");
-	prio_str = strtok(NULL, ":");
+	policy_len = strcspn(string, ":");
+	policy_str = xstrndup(string, policy_len);
 
-	if (prio_str && parse_unsigned(prio_str, 10, &prio) != 0)
+	if (string[policy_len] == ':' &&
+	    parse_unsigned(string + policy_len + 1, 10, &prio) != 0)
 		fatal("invalid process scheduler priority");
 
 	proc_sched = xmalloc(sizeof(*proc_sched));
@@ -632,14 +765,15 @@ parse_proc_schedule(const char *string)
 static void
 parse_io_schedule(const char *string)
 {
-	char *class_str, *prio_str;
+	char *class_str;
+	size_t class_len;
 	int prio = 4;
 
-	class_str = xstrdup(string);
-	class_str = strtok(class_str, ":");
-	prio_str = strtok(NULL, ":");
+	class_len = strcspn(string, ":");
+	class_str = xstrndup(string, class_len);
 
-	if (prio_str && parse_unsigned(prio_str, 10, &prio) != 0)
+	if (string[class_len] == ':' &&
+	    parse_unsigned(string + class_len + 1, 10, &prio) != 0)
 		fatal("invalid IO scheduler priority");
 
 	io_sched = xmalloc(sizeof(*io_sched));
@@ -785,12 +919,16 @@ set_action(enum action_code new_action)
 	if (action == new_action)
 		return;
 
-	if (action != action_none)
+	if (action != ACTION_NONE)
 		badusage("only one command can be specified");
 
 	action = new_action;
 }
 
+#define OPT_PID		500
+#define OPT_PPID	501
+#define OPT_RM_PIDFILE	502
+
 static void
 parse_options(int argc, char * const *argv)
 {
@@ -803,6 +941,8 @@ parse_options(int argc, char * const *argv)
 		{ "startas",	  1, NULL, 'a'},
 		{ "name",	  1, NULL, 'n'},
 		{ "oknodo",	  0, NULL, 'o'},
+		{ "pid",	  1, NULL, OPT_PID},
+		{ "ppid",	  1, NULL, OPT_PPID},
 		{ "pidfile",	  1, NULL, 'p'},
 		{ "quiet",	  0, NULL, 'q'},
 		{ "signal",	  1, NULL, 's'},
@@ -820,15 +960,19 @@ parse_options(int argc, char * const *argv)
 		{ "background",	  0, NULL, 'b'},
 		{ "no-close",	  0, NULL, 'C'},
 		{ "make-pidfile", 0, NULL, 'm'},
+		{ "remove-pidfile", 0, NULL, OPT_RM_PIDFILE},
 		{ "retry",	  1, NULL, 'R'},
 		{ "chdir",	  1, NULL, 'd'},
 		{ NULL,		  0, NULL, 0  }
 	};
+	const char *pid_str = NULL;
+	const char *ppid_str = NULL;
 	const char *umask_str = NULL;
 	const char *signal_str = NULL;
 	const char *schedule_str = NULL;
 	const char *proc_schedule_str = NULL;
 	const char *io_schedule_str = NULL;
+	size_t changeuser_len;
 	int c;
 
 	for (;;) {
@@ -842,13 +986,13 @@ parse_options(int argc, char * const *argv)
 			usage();
 			exit(0);
 		case 'K':  /* --stop */
-			set_action(action_stop);
+			set_action(ACTION_STOP);
 			break;
 		case 'S':  /* --start */
-			set_action(action_start);
+			set_action(ACTION_START);
 			break;
 		case 'T':  /* --status */
-			set_action(action_status);
+			set_action(ACTION_STATUS);
 			break;
 		case 'V':  /* --version */
 			do_version();
@@ -862,6 +1006,12 @@ parse_options(int argc, char * const *argv)
 		case 'o':  /* --oknodo */
 			exitnodo = 0;
 			break;
+		case OPT_PID: /* --pid <pid> */
+			pid_str = optarg;
+			break;
+		case OPT_PPID: /* --ppid <ppid> */
+			ppid_str = optarg;
+			break;
 		case 'p':  /* --pidfile <pid-file> */
 			pidfile = optarg;
 			break;
@@ -886,9 +1036,13 @@ parse_options(int argc, char * const *argv)
 		case 'c':  /* --chuid <username>|<uid> */
 			/* We copy the string just in case we need the
 			 * argument later. */
-			changeuser = xstrdup(optarg);
-			changeuser = strtok(changeuser, ":");
-			changegroup = strtok(NULL, ":");
+			changeuser_len = strcspn(optarg, ":");
+			changeuser = xstrndup(optarg, changeuser_len);
+			if (optarg[changeuser_len] == ':') {
+				if (optarg[changeuser_len + 1] == '\0')
+					fatal("missing group name");
+				changegroup = optarg + changeuser_len + 1;
+			}
 			break;
 		case 'g':  /* --group <group>|<gid> */
 			changegroup = optarg;
@@ -917,6 +1071,9 @@ parse_options(int argc, char * const *argv)
 		case 'm':  /* --make-pidfile */
 			mpidfile = true;
 			break;
+		case OPT_RM_PIDFILE: /* --remove-pidfile */
+			rpidfile = true;
+			break;
 		case 'R':  /* --retry <schedule>|<timeout> */
 			schedule_str = optarg;
 			break;
@@ -929,6 +1086,16 @@ parse_options(int argc, char * const *argv)
 		}
 	}
 
+	if (pid_str != NULL) {
+		if (parse_pid(pid_str, &match_pid) != 0)
+			badusage("pid value must be a number greater than 0");
+	}
+
+	if (ppid_str != NULL) {
+		if (parse_pid(ppid_str, &match_ppid) != 0)
+			badusage("ppid value must be a number greater than 0");
+	}
+
 	if (signal_str != NULL) {
 		if (parse_signal(signal_str, &signal_nr) != 0)
 			badusage("signal value must be numeric or name"
@@ -950,11 +1117,12 @@ parse_options(int argc, char * const *argv)
 			badusage("umask value must be a positive number");
 	}
 
-	if (action == action_none)
+	if (action == ACTION_NONE)
 		badusage("need one of --start or --stop or --status");
 
-	if (!execname && !pidfile && !userspec && !cmdname)
-		badusage("need at least one of --exec, --pidfile, --user or --name");
+	if (!execname && !pid_str && !ppid_str && !pidfile && !userspec &&
+	    !cmdname)
+		badusage("need at least one of --exec, --pid, --ppid, --pidfile, --user or --name");
 
 #ifdef PROCESS_NAME_SIZE
 	if (cmdname && strlen(cmdname) > PROCESS_NAME_SIZE)
@@ -966,13 +1134,18 @@ parse_options(int argc, char * const *argv)
 	if (!startas)
 		startas = execname;
 
-	if (action == action_start && !startas)
+	if (action == ACTION_START && !startas)
 		badusage("--start needs --exec or --startas");
 
 	if (mpidfile && pidfile == NULL)
 		badusage("--make-pidfile requires --pidfile");
+	if (rpidfile && pidfile == NULL)
+		badusage("--remove-pidfile requires --pidfile");
 
-	if (background && action != action_start)
+	if (pid_str && pidfile)
+		badusage("need either --pid of --pidfile, not both");
+
+	if (background && action != ACTION_START)
 		badusage("--background is only relevant with --start");
 
 	if (!close_io && !background)
@@ -1042,6 +1215,40 @@ setup_options(void)
 	}
 }
 
+#if defined(OSLinux)
+static const char *
+proc_status_field(pid_t pid, const char *field)
+{
+	static char *line = NULL;
+	static size_t line_size = 0;
+
+	FILE *fp;
+	char filename[32];
+	char *value = NULL;
+	ssize_t line_len;
+	size_t field_len = strlen(field);
+
+	sprintf(filename, "/proc/%d/status", pid);
+	fp = fopen(filename, "r");
+	if (!fp)
+		return NULL;
+	while ((line_len = getline(&line, &line_size, fp)) >= 0) {
+		if (strncasecmp(line, field, field_len) == 0) {
+			line[line_len - 1] = '\0';
+
+			value = line + field_len;
+			while (isspace(*value))
+				value++;
+
+			break;
+		}
+	}
+	fclose(fp);
+
+	return value;
+}
+#endif
+
 #if defined(OSHurd)
 static void
 init_procset(void)
@@ -1081,6 +1288,36 @@ get_proc_stat(pid_t pid, ps_flags_t flags)
 
 	return ps;
 }
+#elif defined(HAVE_KVM_H)
+static kvm_t *
+ssd_kvm_open(void)
+{
+	kvm_t *kd;
+	char errbuf[_POSIX2_LINE_MAX];
+
+	kd = kvm_openfiles(NULL, KVM_MEMFILE, NULL, O_RDONLY, errbuf);
+	if (kd == NULL)
+		errx(1, "%s", errbuf);
+
+	return kd;
+}
+
+static struct kinfo_proc *
+ssd_kvm_get_procs(kvm_t *kd, int op, int arg, int *count)
+{
+	struct kinfo_proc *kp;
+	int lcount;
+
+	if (count == NULL)
+		count = &lcount;
+	*count = 0;
+
+	kp = kvm_getprocs(kd, op, arg, count);
+	if (kp == NULL && errno != ESRCH)
+		errx(1, "%s", kvm_geterr(kd));
+
+	return kp;
+}
 #endif
 
 #if defined(OSLinux)
@@ -1088,21 +1325,29 @@ static bool
 pid_is_exec(pid_t pid, const struct stat *esb)
 {
 	char lname[32];
-	char lcontents[_POSIX_PATH_MAX];
+	char lcontents[_POSIX_PATH_MAX + 1];
+	char *filename;
 	const char deleted[] = " (deleted)";
 	int nread;
 	struct stat sb;
 
 	sprintf(lname, "/proc/%d/exe", pid);
-	nread = readlink(lname, lcontents, sizeof(lcontents));
+	nread = readlink(lname, lcontents, sizeof(lcontents) - 1);
 	if (nread == -1)
 		return false;
 
-	lcontents[nread] = '\0';
-	if (strcmp(lcontents + nread - strlen(deleted), deleted) == 0)
-		lcontents[nread - strlen(deleted)] = '\0';
+	filename = lcontents;
+	filename[nread] = '\0';
 
-	if (stat(lcontents, &sb) != 0)
+	/* OpenVZ kernels contain a bogus patch that instead of appending,
+	 * prepends the deleted marker. Workaround those. Otherwise handle
+	 * the normal appended marker. */
+	if (strncmp(filename, deleted, strlen(deleted)) == 0)
+		filename += strlen(deleted);
+	else if (strcmp(filename + nread - strlen(deleted), deleted) == 0)
+		filename[nread - strlen(deleted)] = '\0';
+
+	if (stat(filename, &sb) != 0)
 		return false;
 
 	return (sb.st_dev == esb->st_dev && sb.st_ino == esb->st_ino);
@@ -1137,24 +1382,50 @@ pid_is_exec(pid_t pid, const struct stat *esb)
 	return ((dev_t)pst.pst_text.psf_fsid.psfs_id == esb->st_dev &&
 	        (ino_t)pst.pst_text.psf_fileid == esb->st_ino);
 }
+#elif defined(OSFreeBSD)
+static bool
+pid_is_exec(pid_t pid, const struct stat *esb)
+{
+	struct stat sb;
+	int error, name[4];
+	size_t len;
+	char pathname[PATH_MAX];
+
+	name[0] = CTL_KERN;
+	name[1] = KERN_PROC;
+	name[2] = KERN_PROC_PATHNAME;
+	name[3] = pid;
+	len = sizeof(pathname);
+
+	error = sysctl(name, 4, pathname, &len, NULL, 0);
+	if (error != 0 && errno != ESRCH)
+		return false;
+	if (len == 0)
+		pathname[0] = '\0';
+
+	if (stat(pathname, &sb) != 0)
+		return false;
+
+	return (sb.st_dev == esb->st_dev && sb.st_ino == esb->st_ino);
+}
 #elif defined(HAVE_KVM_H)
 static bool
 pid_is_exec(pid_t pid, const struct stat *esb)
 {
 	kvm_t *kd;
-	int nentries, argv_len = 0;
+	int argv_len = 0;
 	struct kinfo_proc *kp;
 	struct stat sb;
-	char errbuf[_POSIX2_LINE_MAX], buf[_POSIX2_LINE_MAX];
+	char buf[_POSIX2_LINE_MAX];
 	char **pid_argv_p;
 	char *start_argv_0_p, *end_argv_0_p;
+	bool res = false;
 
-	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-	if (kd == NULL)
-		errx(1, "%s", errbuf);
-	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	kd = ssd_kvm_open();
+	kp = ssd_kvm_get_procs(kd, KERN_PROC_PID, pid, NULL);
 	if (kp == NULL)
-		errx(1, "%s", kvm_geterr(kd));
+		goto cleanup;
+
 	pid_argv_p = kvm_getargv(kd, kp, argv_len);
 	if (pid_argv_p == NULL)
 		errx(1, "%s", kvm_geterr(kd));
@@ -1178,9 +1449,91 @@ pid_is_exec(pid_t pid, const struct stat *esb)
 	}
 
 	if (stat(start_argv_0_p, &sb) != 0)
+		goto cleanup;
+
+	res = (sb.st_dev == esb->st_dev && sb.st_ino == esb->st_ino);
+
+cleanup:
+	kvm_close(kd);
+
+	return res;
+}
+#endif
+
+#if defined(OSLinux)
+static bool
+pid_is_child(pid_t pid, pid_t ppid)
+{
+	const char *ppid_str;
+	pid_t proc_ppid;
+	int rc;
+
+	ppid_str = proc_status_field(pid, "PPid:");
+	if (ppid_str == NULL)
 		return false;
 
-	return (sb.st_dev == esb->st_dev && sb.st_ino == esb->st_ino);
+	rc = parse_pid(ppid_str, &proc_ppid);
+	if (rc < 0)
+		return false;
+
+	return proc_ppid == ppid;
+}
+#elif defined(OSHurd)
+static bool
+pid_is_child(pid_t pid, pid_t ppid)
+{
+	struct proc_stat *ps;
+	struct procinfo *pi;
+
+	ps = get_proc_stat(pid, PSTAT_PROC_INFO);
+	if (ps == NULL)
+		return false;
+
+	pi = proc_stat_proc_info(ps);
+
+	return pi->ppid == ppid;
+}
+#elif defined(OShpux)
+static bool
+pid_is_child(pid_t pid, pid_t ppid)
+{
+	struct pst_status pst;
+
+	if (pstat_getproc(&pst, sizeof(pst), (size_t)0, (int)pid) < 0)
+		return false;
+
+	return pst.pst_ppid == ppid;
+}
+#elif defined(HAVE_KVM_H)
+static bool
+pid_is_child(pid_t pid, pid_t ppid)
+{
+	kvm_t *kd;
+	struct kinfo_proc *kp;
+	pid_t proc_ppid;
+	bool res = false;
+
+	kd = ssd_kvm_open();
+	kp = ssd_kvm_get_procs(kd, KERN_PROC_PID, pid, NULL);
+	if (kp == NULL)
+		goto cleanup;
+
+#if defined(OSFreeBSD)
+	proc_ppid = kp->ki_ppid;
+#elif defined(OSOpenBSD)
+	proc_ppid = kp->p_ppid;
+#elif defined(OSDragonFlyBSD)
+	proc_ppid = kp->kp_ppid;
+#else
+	proc_ppid = kp->kp_proc.p_ppid;
+#endif
+
+	res = (proc_ppid == ppid);
+
+cleanup:
+	kvm_close(kd);
+
+	return res;
 }
 #endif
 
@@ -1220,23 +1573,35 @@ static bool
 pid_is_user(pid_t pid, uid_t uid)
 {
 	kvm_t *kd;
-	int nentries; /* Value not used. */
 	uid_t proc_uid;
 	struct kinfo_proc *kp;
-	char errbuf[_POSIX2_LINE_MAX];
+	bool res = false;
 
-	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-	if (kd == NULL)
-		errx(1, "%s", errbuf);
-	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	kd = ssd_kvm_open();
+	kp = ssd_kvm_get_procs(kd, KERN_PROC_PID, pid, NULL);
 	if (kp == NULL)
-		errx(1, "%s", kvm_geterr(kd));
+		goto cleanup;
+
+#if defined(OSFreeBSD)
+	proc_uid = kp->ki_ruid;
+#elif defined(OSOpenBSD)
+	proc_uid = kp->p_ruid;
+#elif defined(OSDragonFlyBSD)
+	proc_uid = kp->kp_ruid;
+#else
 	if (kp->kp_proc.p_cred)
 		kvm_read(kd, (u_long)&(kp->kp_proc.p_cred->p_ruid),
 		         &proc_uid, sizeof(uid_t));
 	else
-		return false;
-	return (proc_uid == (uid_t)uid);
+		goto cleanup;
+#endif
+
+	res = (proc_uid == (uid_t)uid);
+
+cleanup:
+	kvm_close(kd);
+
+	return res;
 }
 #endif
 
@@ -1244,25 +1609,13 @@ pid_is_user(pid_t pid, uid_t uid)
 static bool
 pid_is_cmd(pid_t pid, const char *name)
 {
-	char buf[32];
-	FILE *f;
-	int c;
+	const char *comm;
 
-	sprintf(buf, "/proc/%d/stat", pid);
-	f = fopen(buf, "r");
-	if (!f)
+	comm = proc_status_field(pid, "Name:");
+	if (comm == NULL)
 		return false;
-	while ((c = getc(f)) != EOF && c != '(')
-		;
-	if (c != '(') {
-		fclose(f);
-		return false;
-	}
-	/* This hopefully handles command names containing ‘)’. */
-	while ((c = getc(f)) != EOF && c == *name)
-		name++;
-	fclose(f);
-	return (c == ')' && *name == '\0');
+
+	return strcmp(comm, name) == 0;
 }
 #elif defined(OSHurd)
 static bool
@@ -1310,20 +1663,31 @@ static bool
 pid_is_cmd(pid_t pid, const char *name)
 {
 	kvm_t *kd;
-	int nentries;
 	struct kinfo_proc *kp;
-	char errbuf[_POSIX2_LINE_MAX], *process_name;
+	char *process_name;
+	bool res = false;
 
-	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-	if (kd == NULL)
-		errx(1, "%s", errbuf);
-	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	kd = ssd_kvm_open();
+	kp = ssd_kvm_get_procs(kd, KERN_PROC_PID, pid, NULL);
 	if (kp == NULL)
-		errx(1, "%s", kvm_geterr(kd));
-	process_name = (&kp->kp_proc)->p_comm;
-	if (strlen(name) != strlen(process_name))
-		return false;
-	return (strcmp(name, process_name) == 0);
+		goto cleanup;
+
+#if defined(OSFreeBSD)
+	process_name = kp->ki_comm;
+#elif defined(OSOpenBSD)
+	process_name = kp->p_comm;
+#elif defined(OSDragonFlyBSD)
+	process_name = kp->kp_comm;
+#else
+	process_name = kp->kp_proc.p_comm;
+#endif
+
+	res = (strcmp(name, process_name) == 0);
+
+cleanup:
+	kvm_close(kd);
+
+	return res;
 }
 #endif
 
@@ -1350,17 +1714,19 @@ static enum status_code
 pid_check(pid_t pid)
 {
 	if (execname && !pid_is_exec(pid, &exec_stat))
-		return status_dead;
+		return STATUS_DEAD;
+	if (match_ppid > 0 && !pid_is_child(pid, match_ppid))
+		return STATUS_DEAD;
 	if (userspec && !pid_is_user(pid, user_id))
-		return status_dead;
+		return STATUS_DEAD;
 	if (cmdname && !pid_is_cmd(pid, cmdname))
-		return status_dead;
-	if (action != action_stop && !pid_is_running(pid))
-		return status_dead;
+		return STATUS_DEAD;
+	if (action != ACTION_STOP && !pid_is_running(pid))
+		return STATUS_DEAD;
 
 	pid_list_push(&found, pid);
 
-	return status_ok;
+	return STATUS_OK;
 }
 
 static enum status_code
@@ -1379,15 +1745,15 @@ do_pidfile(const char *name)
 		if (fscanf(f, "%d", &pid) == 1)
 			pid_status = pid_check(pid);
 		else
-			pid_status = status_unknown;
+			pid_status = STATUS_UNKNOWN;
 		fclose(f);
 
-		if (pid_status == status_dead)
-			return status_dead_pidfile;
+		if (pid_status == STATUS_DEAD)
+			return STATUS_DEAD_PIDFILE;
 		else
 			return pid_status;
 	} else if (errno == ENOENT)
-		return status_dead;
+		return STATUS_DEAD;
 	else
 		fatal("unable to open pidfile %s", name);
 }
@@ -1400,7 +1766,7 @@ do_procinit(void)
 	struct dirent *entry;
 	int foundany;
 	pid_t pid;
-	enum status_code prog_status = status_dead;
+	enum status_code prog_status = STATUS_DEAD;
 
 	procdir = opendir("/proc");
 	if (!procdir)
@@ -1441,9 +1807,9 @@ do_procinit(void)
 	proc_stat_list_for_each(procset, check_proc_stat);
 
 	if (found)
-		return status_ok;
+		return STATUS_OK;
 	else
-		return status_dead;
+		return STATUS_DEAD;
 }
 #elif defined(OShpux)
 static enum status_code
@@ -1452,7 +1818,7 @@ do_procinit(void)
 	struct pst_status pst[10];
 	int i, count;
 	int idx = 0;
-	enum status_code prog_status = status_dead;
+	enum status_code prog_status = STATUS_DEAD;
 
 	while ((count = pstat_getproc(pst, sizeof(pst[0]), 10, idx)) > 0) {
 		enum status_code pid_status;
@@ -1471,8 +1837,36 @@ do_procinit(void)
 static enum status_code
 do_procinit(void)
 {
-	/* Nothing to do. */
-	return status_unknown;
+	kvm_t *kd;
+	int nentries, i;
+	struct kinfo_proc *kp;
+	enum status_code prog_status = STATUS_DEAD;
+
+	kd = ssd_kvm_open();
+	kp = ssd_kvm_get_procs(kd, KERN_PROC_ALL, 0, &nentries);
+
+	for (i = 0; i < nentries; i++) {
+		enum status_code pid_status;
+		pid_t pid;
+
+#if defined(OSFreeBSD)
+		pid = kp[i].ki_pid;
+#elif defined(OSOpenBSD)
+		pid = kp[i].p_pid;
+#elif defined(OSDragonFlyBSD)
+		pid = kp[i].kp_pid;
+#else
+		pid = kp[i].kp_proc.p_pid;
+#endif
+
+		pid_status = pid_check(pid);
+		if (pid_status < prog_status)
+			prog_status = pid_status;
+	}
+
+	kvm_close(kd);
+
+	return prog_status;
 }
 #endif
 
@@ -1481,7 +1875,9 @@ do_findprocs(void)
 {
 	pid_list_free(&found);
 
-	if (pidfile)
+	if (match_pid > 0)
+		return pid_check(match_pid);
+	else if (pidfile)
 		return do_pidfile(pidfile);
 	else
 		return do_procinit();
@@ -1532,6 +1928,9 @@ do_start(int argc, char **argv)
 	if (background)
 		/* Ok, we need to detach this process. */
 		daemonize();
+	else if (mpidfile && pidfile != NULL)
+		/* User wants _us_ to make the pidfile, but detach themself! */
+		write_pidfile(pidfile, getpid());
 	if (background && close_io) {
 		devnull_fd = open("/dev/null", O_RDWR);
 		if (devnull_fd < 0)
@@ -1548,9 +1947,6 @@ do_start(int argc, char **argv)
 		set_io_schedule(io_sched);
 	if (umask_value >= 0)
 		umask(umask_value);
-	if (mpidfile && pidfile != NULL)
-		/* User wants _us_ to make the pidfile. */
-		write_pidfile(pidfile, getpid());
 	if (changeroot != NULL) {
 		if (chdir(changeroot) < 0)
 			fatal("unable to chdir() to %s", changeroot);
@@ -1581,10 +1977,6 @@ do_start(int argc, char **argv)
 				fatal("unable to set uid to %s", changeuser);
 	}
 
-	/* Set a default umask for dumb programs. */
-	if (background && umask_value < 0)
-		umask(022);
-
 	if (background && close_io) {
 		int i;
 
@@ -1680,45 +2072,45 @@ set_what_stop(const char *str)
 static bool
 do_stop_timeout(int timeout, int *n_killed, int *n_notkilled)
 {
-	struct timeval stopat, before, after, interval, maxinterval;
-	int r, ratio;
+	struct timespec stopat, before, after, interval, maxinterval;
+	int rc, ratio;
 
-	xgettimeofday(&stopat);
+	timespec_gettime(&stopat);
 	stopat.tv_sec += timeout;
 	ratio = 1;
 	for (;;) {
-		xgettimeofday(&before);
-		if (timercmp(&before, &stopat, >))
+		timespec_gettime(&before);
+		if (timespec_cmp(&before, &stopat, >))
 			return false;
 
 		do_stop(0, n_killed, n_notkilled);
 		if (!*n_killed)
 			return true;
 
-		xgettimeofday(&after);
+		timespec_gettime(&after);
 
-		if (!timercmp(&after, &stopat, <))
+		if (!timespec_cmp(&after, &stopat, <))
 			return false;
 
 		if (ratio < 10)
 			ratio++;
 
-		timersub(&stopat, &after, &maxinterval);
-		timersub(&after, &before, &interval);
-		tmul(&interval, ratio);
+		timespec_sub(&stopat, &after, &maxinterval);
+		timespec_sub(&after, &before, &interval);
+		timespec_mul(&interval, ratio);
 
-		if (interval.tv_sec < 0 || interval.tv_usec < 0)
-			interval.tv_sec = interval.tv_usec = 0;
+		if (interval.tv_sec < 0 || interval.tv_nsec < 0)
+			interval.tv_sec = interval.tv_nsec = 0;
 
-		if (timercmp(&interval, &maxinterval, >))
+		if (timespec_cmp(&interval, &maxinterval, >))
 			interval = maxinterval;
 
 		if (interval.tv_sec == 0 &&
-		    interval.tv_usec <= MIN_POLL_INTERVAL)
-			interval.tv_usec = MIN_POLL_INTERVAL;
+		    interval.tv_nsec <= MIN_POLL_INTERVAL)
+			interval.tv_nsec = MIN_POLL_INTERVAL;
 
-		r = select(0, NULL, NULL, NULL, &interval);
-		if (r < 0 && errno != EINTR)
+		rc = pselect(0, NULL, NULL, NULL, &interval, NULL);
+		if (rc < 0 && errno != EINTR)
 			fatal("select() failed for pause");
 	}
 }
@@ -1726,6 +2118,9 @@ do_stop_timeout(int timeout, int *n_killed, int *n_notkilled)
 static int
 finish_stop_schedule(bool anykilled)
 {
+	if (rpidfile && pidfile && !testmode)
+		remove_pidfile(pidfile);
+
 	if (anykilled)
 		return 0;
 
@@ -1755,6 +2150,10 @@ run_stop_schedule(void)
 		set_what_stop(execname);
 	else if (pidfile)
 		sprintf(what_stop, "process in pidfile '%.200s'", pidfile);
+	else if (match_pid > 0)
+		sprintf(what_stop, "process with pid %d", match_pid);
+	else if (match_ppid > 0)
+		sprintf(what_stop, "process(es) with parent pid %d", match_ppid);
 	else if (userspec)
 		sprintf(what_stop, "process(es) owned by '%.200s'", userspec);
 	else
@@ -1818,15 +2217,15 @@ main(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (action == action_start)
+	if (action == ACTION_START)
 		do_start(argc, argv);
 
-	if (action == action_stop) {
+	if (action == ACTION_STOP) {
 		int i = run_stop_schedule();
 		exit(i);
 	}
 
-	if (action == action_status) {
+	if (action == ACTION_STATUS) {
 		enum status_code prog_status;
 
 		prog_status = do_findprocs();
